from compas_plotters import Plotter
from compas.geometry import Vector
from compas.geometry import Point
from compas.geometry import Rotation
from compas.geometry import Translation
from compas.geometry import Scale
from compas.geometry import Frame
from compas.geometry import Line
from compas.geometry import Polygon
from compas_tno.shapes import Shape

from compas.colors import Color

from compas.utilities import rgb_to_hex
import matplotlib.pyplot as plt
import math
from PIL import Image


class TNOPlotter(object):
    """TNO helper to plot forms and shapes

    Parameters
    ----------
    form : FormDiagram, optional
        The form diagram to be plotted, by default None
    shape : Shape, optional
        The shape of the masonry constraining the solution, by default None
    form_base : FormDiagram, optional
        The base form diagram to be visualised if pattern is allowed to move, by default None
    force : ForceDiagram, optional
        The force diagram associates with the form diagram, by default None
    figsize : tuple, optional
        The size of the plot, by default (8, 8)

    Attributes
    ----------

    None

    """

    def __init__(self, form=None,
                 shape=None,
                 form_base=None,
                 force=None,
                 figsize=(8, 8),
                 *args,
                 **kwargs):

        super().__init__(*args, **kwargs)
        self.title = 'Plotter'
        self.app = None
        self._form = form
        self._shape = shape
        self._form_base = form_base
        self._force = force
        self._formartist = None
        self._otherartists = []
        self.scale_forcediagram = 1.0
        self.translation_forcediagram = [0.0, 0.0, 0.0]
        self.settings = {
            'show.thrust': True,
            'show.shape': True,
            'show.reactions': True,
            'show.reactions.emerging': True,
            'show.reactionlabels': True,
            'show.supports': True,
            'show.edges': True,
            'show.faces': False,
            'show.cracks': True,
            'show.vertex.outside': True,
            'show.edge.thickness': True,
            'show.reactions.extended': False,
            'show.reactions.asarrows': True,

            'camera.target': [0, 0, 0],
            'camera.distance': 40,
            'camera.rx': -45,
            'camera.rz': 45,
            'camera.fov': 40,
            'camera.show.grid': True,
            'camera.show.axis': True,

            'size.vertex': 5.0,
            'size.edge.max_thickness': 10.0,
            'size.edge.base_thickness': 1.0,
            'size.edge.normals': 0.5,
            'size.reaction.head_width': 0.1,
            'size.reaction.body_width': 0.01,
            'size.reactionlabel': 20,

            'scale.reactions': 0.05,
            'scale.loads': 0.001,
            'shape.opacity': 0.5,
            'rotated': False,

            'color.edges.form': Color.from_rgb255(255, 0, 0),
            'color.edges.reactions': Color.from_rgb255(170, 170, 170),
            'color.edges.shape': Color.from_rgb255(200, 200, 200),
            'color.edges.form_base': Color.from_rgb255(200, 200, 200),
            'color.edges.independent': Color.from_rgb255(0, 0, 255),
            'color.vertex.supports': Color.from_rgb255(170, 170, 170),
            'color.vertex.extrados': Color.from_rgb255(0, 125, 0),
            'color.vertex.intrados': Color.from_rgb255(0, 0, 255),
            'color.vertex.outside': Color.from_rgb255(200, 200, 200),
            'color.mesh.intrados': Color.from_rgb255(125, 125, 125),
            'color.mesh.extrados': Color.from_rgb255(125, 125, 125),
            'color.mesh.middle': Color.from_rgb255(125, 125, 125),
            'color.mesh.general': Color.from_rgb255(0, 0, 0),

            'tol.forces': 1e-3,
        }
        self.initiate_app(figsize=figsize)

    @property
    def form(self):
        """The FormDiagram of the Plotter."""
        return self._form

    @form.setter
    def form(self, value):
        self._form = value

    @property
    def shape(self):
        """The Shape of the Plotter."""
        return self._shape

    @shape.setter
    def shape(self, value):
        self._shape = value

    @property
    def form_base(self):
        """The form_base property."""
        return self._form_base

    @form_base.setter
    def form_base(self, value):
        self._form_base = value

    @property
    def force(self):
        """The force property."""
        return self._force

    @force.setter
    def force(self, value):
        self._force = value

    @property
    def formartist(self):
        """The formartist property."""
        return self._formartist

    @formartist.setter
    def formartist(self, value):
        if not self._formartist:
            self._formartist = value

    def initiate_app(self, figsize=(8, 8)):
        """Initiate the Plotter with the default camera options.

        Parameters
        ----------
        figsize : tuple, optional
            Size of the plot, by default (8, 8)

        Returns
        -------
        None
            The plotter is updated in place
        """

        self.app = Plotter(figsize=figsize)
        self.app.fontsize = 10

    def show_solution(self):
        """Show the thrust network, with the shape according to the settings.

        Returns
        -------
        None
            The plotter is updated in place
        """

        self.draw_form()
        self.draw_supports()
        self.draw_cracks()
        self.zoom_extents()
        self.show()

    def zoom_extents(self, padding=0.1):
        """Wrapper to extend the objects in the active view.

        Parameters
        ----------
        padding : int, optional
            Additional white space surrounding the object, by default 0.1.

        Returns
        -------
        None
            The plotter is updated in place
        """

        self.app.zoom_extents(padding=padding)

    def show(self, padding=0.1):
        """Display the plotter in the screen.

        Parameters
        ----------
        padding : int, optional
            Additional white space surrounding the object, by default 0.1.

        Returns
        -------
        None
            The plotter is updated in place
        """

        self.app.zoom_extents(padding=padding)
        self.app.show()

    def clear(self):
        """Clear the plotter elements.

        Returns
        -------
        None
            The plotter is updated in place
        """

        self.app.clear()

    def save(self, filepath):
        """Save the plotter window as an image.

        Parameters
        ----------
        filepath : str
            The file path to save the image.

        Returns
        -------
        None
            Picture is saved

        """

        # self.zoom_extents()

        self.app.viewbox = ([10, 30], [-1, 10])
        self.app.axes.set_xlim([10, 30])
        self.app.axes.set_ylim([-1, 10])
        self.app.axes.autoscale_view()

        self.app.save(filepath)

    def add(self, item, **kwargs):
        """Add an item to the plotter.

        Parameters
        ----------
        item : Any
            Item to add

        Returns
        -------
        Artist
            The artist of the added item

        """

        return self.app.add(item, **kwargs)

    def draw_form(self, scale_width=True, edges=None, color=None, **kwargs):
        """Draw the Form Diagram with or without thicknesses of edges scaled as forces in the edges.

        Parameters
        ----------
        scale_width : bool, optional
            If the lines of the form diagram should be scaled with regards to the force carried, by default True
        color : Color, optional
            Color of the edges of the form diagram, this will overwrite the setting 'color.edges.form', by default None
        edges : [tuple], optional
            Edges to consider in the plot, by default None, in which all edges are plotted

        Returns
        -------
        None
            The plotter is updated in place
        """

        base_thick = self.settings['size.edge.base_thickness']
        max_thick = self.settings['size.edge.max_thickness']
        color = color or self.settings['color.edges.form']
        edgecolor = {(u, v): color for u, v in self.form.edges()}

        if scale_width:
            forcedensities = self.form.edges_attribute('q')
            lengths = [self.form.edge_length(u, v) for u, v in self.form.edges()]
            forces = [abs(forcedensities[i] * lengths[i]) for i in range(len(lengths))]
            fmax = max(forces)
            edgewidths = {edge: forces[i] / fmax * max_thick for i, edge in enumerate(self.form.edges())}
        else:
            edgewidths = base_thick
            edgecolor = {(u, v): Color.black() for u, v in self.form.edges()}

        faces = []

        if self.settings['show.edges']:
            edges = edges or list(self.form.edges_where({'_is_edge': True}))

        if self.settings['show.faces']:
            faces = list(self.form.faces())

        formartist = self.app.add(  # change this to a bunch of lines.... Better to edit
            self.form,              # add cull_negative option to cull the thrust line below a value...
            # vertices=[],
            edges=edges,
            faces=faces,
            edgewidth=edgewidths,
            edgecolor=edgecolor,
            show_vertices=False,
            **kwargs
        )

        self.formartist = formartist

        # artist = self.app.add(
        #     self.form,
        #     edges=edges,
        #     edgewidth=width,
        #     edgetext=text,
        #     edgecolor=color,
        #     show_vertices=False,
        #     show_faces=False
        # )
        # self.formartist = artist

    def draw_cracks(self, points=None, **kwargs):
        """Adds to the basic plot, the cracks which are the points of the mesh that touch intrados or extrados.

        Points
        ------
        points : list, optional
            The list with the cracks to consider

        Returns
        -------
        None
            The plotter is updated in place
        """

        tol = self.settings['tol.forces']
        color_intra = self.settings['color.vertex.intrados']
        color_extra = self.settings['color.vertex.extrados']
        color_outside = self.settings['color.vertex.outside']

        cracks = []
        points = points or list(self.form.vertices())

        for key in points:
            x, y, z = self.form.vertex_coordinates(key)
            if self.settings['rotated']:
                x, z, y = self.form.vertex_coordinates(key)
            lb = self.form.vertex_attribute(key, 'lb')
            ub = self.form.vertex_attribute(key, 'ub')
            if not lb:
                continue
            if not ub:
                continue
            if self.settings['show.cracks']:
                if abs(ub - z) < tol:
                    cracks.append({
                        'key': key,
                        'xyz': [x, y, z],
                        'color': color_extra
                    })
                    continue
                elif abs(lb - z) < tol:
                    cracks.append({
                        'key': key,
                        'xyz': [x, y, z],
                        'color': color_intra
                    })
                    continue
            if self.settings['show.vertex.outside']:
                if z - ub > tol or lb - z > tol:
                    cracks.append({
                        'key': key,
                        'xyz': [x, y, z],
                        'color': color_outside
                    })
                    continue

        if self.settings['show.reactions.extended']:
            for key in self.form.vertices_where({'is_fixed': True}):
                x, y, z = self.form.vertex_coordinates(key)
                if self.settings['rotated']:
                    x, z, y = self.form.vertex_coordinates(key)
                b = self.form.vertex_attribute(key, 'b')
                if z > tol:
                    reaction_scale = z/self.form.vertex_attribute(key, '_rz')
                else:
                    continue
                rx = self.form.vertex_attribute(key, '_rx') * reaction_scale
                ry = self.form.vertex_attribute(key, '_ry') * reaction_scale
                rz = self.form.vertex_attribute(key, '_rz') * reaction_scale
                if abs(abs(rx) - b[0]) < tol:
                    cracks.append({
                        'key': key,
                        'xyz': [x - rx, y - ry, z - rz],
                        'color': color_extra
                    })

        for point in cracks:
            x, y, z = point['xyz']
            if self.settings['rotated']:
                x, z, y = point['xyz']
            color = point['color']
            pt = Point(x, y, z)
            pointartist = self.app.add(pt, facecolor=color, size=self.settings['size.vertex'])
            self._otherartists.append(pointartist)

    def draw_supports(self, color=None, size=None):
        """Add the supports as points in the mesh.

        Parameters
        ----------
        color : Color, optional
            Color of the supports of the form diagram, this will overwrite the setting 'color.vertex.supports', by default None
        size : float, optional
            Size of the supports, this will overwrite the setting 'size.vertex', by default None

        Returns
        -------
        None
            The plotter is updated in place
        """

        if self.settings['show.supports']:
            supportcolor = color or self.settings['color.vertex.supports']  # _norm(self.settings['color.vertex.supports'])
            size = size or self.settings['size.vertex']
            rollercolor = Color.from_rgb255(255, 165, 0)

            for key in self.form.vertices_where({'is_fixed': True}):
                x, y, z = self.form.vertex_coordinates(key)
                pt = Point(x, y, z)
                pointartist = self.app.add(pt, facecolor=supportcolor, size=size)
                self._otherartists.append(pointartist)

            rollers = list(self.form.vertices_where({'rol_x': True})) + list(self.form.vertices_where({'rol_y': True}))
            for key in rollers:
                x, y, z = self.form.vertex_coordinates(key)
                pt = Point(x, y, z)
                pointartist = self.app.add(pt, facecolor=rollercolor, size=self.settings['size.vertex'])
                self._otherartists.append(pointartist)

    def draw_reactions(self, scale_width=True):
        """Add to the plots the vector of the reaction forces.

        Parameters
        ----------
        scale_width : bool, optional
            If edges should be scaled with the force, by default True

        Returns
        -------
        None
            The plotter is updated in place
        """

        if self.settings['show.reactions']:
            reaction_scale = self.settings['scale.reactions']

            for key in self.form.vertices_where({'is_fixed': True}):
                x, y, z = self.form.vertex_coordinates(key)
                if self.settings['show.reactions.extended']:  # assume rotated
                    if y > 0:
                        reaction_scale = y/self.form.vertex_attribute(key, '_rz')
                    else:
                        reaction_scale = 0.0
                rx = self.form.vertex_attribute(key, '_rx') * reaction_scale
                ry = self.form.vertex_attribute(key, '_ry') * reaction_scale
                rz = self.form.vertex_attribute(key, '_rz') * reaction_scale
                norm = math.sqrt(self.form.vertex_attribute(key, '_rx')**2 + self.form.vertex_attribute(key, '_ry')**2 + self.form.vertex_attribute(key, '_rz')**2)
                if self.settings['rotated']:
                    rx, ry, rz = rx, rz, ry
                if self.settings['show.reactions.emerging']:
                    r = Vector(-rx, -ry, -rz)
                    pt = Point(x, y, z)
                else:
                    r = Vector(rx, ry, rz)
                    pt = Point(x - rx, y - ry, z - rz)
                if self.settings['show.reactions.asarrows']:
                    self.app.add(r, point=pt, color=self.settings['color.edges.reactions'])
                else:
                    pt1 = Point(x, y, z)
                    pt2 = Point(x - rx, y - ry, z - rz)
                    line = Line(pt1, pt2)
                    max_f = max([abs(self.form.edge_attribute(edge, 'q') * self.form.edge_length(*edge)) for edge in self.form.edges()])
                    if scale_width:
                        width = max_f/norm*self.settings['size.edge.max_thickness']
                    else:
                        width = self.settings['size.edge.base_thickness']
                    self.app.add(line,
                                 draw_as_segment=True,
                                 color=self.settings['color.edges.form'],
                                 linewidth=width)

    def draw_vector(self, vector=None, base=None):
        """Helper to add vectors to the plotter.

        Parameters
        ----------
        vectors : Vector, optional
            Vector to plot, by default None
        bases : Point, optional
            Point with the location of the base of the vector, by default None

        Returns
        -------
        None
            The plotter is updated in place
        """

        if not vector or not base:
            raise ValueError()

        vectorartist = self.app.add(vector, point=base)
        self._otherartists.append(vectorartist)

    def draw_lines(self, lines=[]):
        """Helper to add vectors to the plotter.

        Parameters
        ----------
        lines : list, optional
            The list of lines to plot, by default []

        Returns
        -------
        None
            The plotter is updated in place
        """

        for line in lines:
            linecompas = Line(line[0], line[1])
            self.app.add(linecompas, draw_as_segment=True)

    def draw_mesh(self, mesh=None, show_edges=True, show_vertices=False, show_faces=False, **kwargs):
        """Initiate the Plotter with the default camera options.

        Parameters
        ----------
        mesh : Mesh, optional
            Mesh to plot, by default None, which means that the form diagram is added
        show_edges : bool, optional
            If edges are shown, by default True
        show_vertices : bool, optional
            If vertices are shown, by default False
        show_faces : bool, optional
            If faces are shown, by default False

        Returns
        -------
        None
            The plotter is updated in place.
        """

        if not mesh:
            mesh = self.form
        self.app.add(mesh,
                     show_edges=show_edges,
                     show_vertices=show_vertices,
                     show_faces=show_faces,
                     **kwargs
                     )

    def draw_form_xz(self, scale_width=True, edges=None, **kwargs):
        """Plot the form diagram rotated 90 degrees

        Parameters
        ----------
        scale_width : bool, optional
            If edges should be scaled with the force, by default True
        edges : [tuple], optional
            Edges to consider in the plot, by default None, in which all edges are plotted

        Returns
        -------
        None
            The plotter is updated in place.
        """

        axis = Vector(1.0, 0, 0)
        self.form = self.form.transformed(Rotation.from_axis_and_angle(axis, -math.pi/2))
        self.settings['rotated'] = True

        self.draw_form(scale_width=scale_width, edges=edges, **kwargs)

    def draw_shape(self, update_from_parameters=True, **kwargs):
        """Adds the shape to the plot.

        Parameters
        ----------
        update_from_parameters : bool, optional
            Whether or not should look for shapes having special cases such as "arch" and "dome" to adapt.

        Returns
        -------
        None
            The plotter is updated in place.
        """

        if not self.shape:
            self.shape = Shape.from_formdiagram_and_attributes(self.form)

        intrados = self.shape.intrados
        extrados = self.shape.extrados

        if update_from_parameters:
            datashape = self.shape.datashape.copy()
            if datashape['type'] in ['arch']:
                stereotomy = False
                close_bottom = True
                total_nodes = datashape['discretisation']
                for key in kwargs:
                    if key == 'stereotomy':
                        stereotomy = kwargs[key]
                    if key == 'close_bottom':
                        close_bottom = kwargs[key]
                    if key == 'blocks':
                        total_nodes = kwargs[key]
                self.draw_arch_lines(H=datashape['H'],
                                     L=datashape['L'],
                                     x0=datashape['x0'],
                                     thk=datashape['thk'],
                                     total_nodes=total_nodes,
                                     stereotomy=stereotomy,
                                     close_bottom=close_bottom)
                return

        self.app.add(
            intrados,
            opacity=self.settings['shape.opacity'],
            edgecolor={edge: self.settings['color.edges.shape'] for edge in intrados.edges()},
            show_vertices=False
        )

        self.app.add(
            extrados,
            opacity=self.settings['shape.opacity'],
            edgecolor={edge: self.settings['color.edges.shape'] for edge in extrados.edges()},
            show_vertices=False
        )

    def draw_shape_xz(self, update_from_parameters=True, **kwargs):
        """Adds the shape to the plot rotated 90 degrees.

        Parameters
        ----------
        update_from_parameters : bool, optional
            Whether or not should look for shapes having special cases such as "arch" and "dome" to adapt.

        Returns
        -------
        None
            The plotter is updated in place
        """

        if not self.shape:
            self.shape = Shape.from_formdiagram_and_attributes(self.form)

        axis = Vector(1.0, 0, 0)
        self.shape.intrados = self.shape.intrados.transformed(Rotation.from_axis_and_angle(axis, -math.pi/2))
        self.shape.extrados = self.shape.extrados.transformed(Rotation.from_axis_and_angle(axis, -math.pi/2))
        self.settings['rotated'] = True

        self.draw_shape(update_from_parameters=update_from_parameters, **kwargs)

    def draw_base_form(self, form_base=None):
        """Adds to the plot the base mesh which is the mesh before the nodes moved horizontally.

        Parameters
        ----------
        form_base : :class:`~compas_tno.diagrams.FormDiagram`, optional
            The base to plot, by default None.

        Returns
        -------
        None
            The plotter is updated in place
        """

        if form_base:
            self.form_base = form_base

        if not self.form_base:
            return

        edgecolors = {edge: self.settings['color.edges.form_base'] for edge in self.form_base.edges()}

        if self.settings['show.edges']:
            edges = list(self.form_base.edges_where({'_is_edge': True}))

        self.app.add(
            self.form_base,
            edges=edges,
            opacity=0.5,
            edgecolor=edgecolors,
            show_vertices=False,
            show_faces=False
        )

    def draw_constraints(self):
        """ Draw the constraints applied.

        Returns
        -------
        None
            The Viewer object is modified in place
        """

        draw_xy_constraints = True

        if draw_xy_constraints:
            for key in self.form.vertices():
                x, y, _ = self.form.vertex_coordinates(key)
                xmin = self.form.vertex_attribute(key, 'xmin')
                xmax = self.form.vertex_attribute(key, 'xmax')
                ymin = self.form.vertex_attribute(key, 'ymin')
                ymax = self.form.vertex_attribute(key, 'ymax')
                if xmin:
                    p0 = [xmin, ymin]
                    p1 = [xmax, ymin]
                    p2 = [xmax, ymax]
                    p3 = [xmin, ymax]
                    lines = [Line(p0, p1), Line(p1, p2), Line(p2, p3), Line(p3, p0)]
                    for line in lines:
                        self.app.add(line, draw_as_segment=True, color=self.settings['color.edges.form_base'])

    def draw_form_independents(self, color=None, show_text=True, **kwargs):
        """Draw the form diagram with highlight in the independent edges.

        Parameters
        ----------
        show_text : bool, optional
            If text should be added to the independent edges, by default True
        color : Color, optional
            Color of the independent edges of the form diagram, this will overwrite the setting 'color.edges.independent', by default None

        Returns
        -------
        None
            The plotter is updated in place
        """

        edges = list(self.form.edges_where({'_is_edge': True}))
        base_thick = self.settings['size.edge.base_thickness']
        colorinds = color or self.settings['color.edges.independent']

        color = {}
        width = {}
        text = {}
        i = 0
        for edge in edges:
            if self.form.edge_attribute(edge, 'is_ind'):
                color[edge] = colorinds
                width[edge] = base_thick * 3.0
                if show_text:
                    text[edge] = str(i)
                i = i+1
            else:
                color[edge] = (0, 0, 0)
                width[edge] = base_thick

        artist = self.app.add(
            self.form,
            edges=edges,
            edgewidth=width,
            edgetext=text,
            edgecolor=color,
            show_vertices=False,
            show_faces=False
        )
        self.formartist = artist

    def draw_vertexlabels(self, label=None, fontsize=12, **kwargs):
        """Draw a specific label in the vertices

        Parameters
        ----------
        label : str, optional
            The vertex label to be shown, by default None, in which case the keys are displayed.
        fontsize : int, optional
            Font size, by default 12

        """

        if not self.formartist:
            return

        text = {}

        if not label:
            for key in self.form.vertices():
                text[key] = key
        elif isinstance(label, str):
            for key in self.form.vertices():
                text[key] = self.form.vertex_attribute(key, label)
        else:
            return

        self.formartist.draw_vertexlabels(text=text)

    # def draw_vertexlabels(self, text=None, fontsize=12):
    #     """Draw Labels to the vertices

    #     Parameters
    #     ----------
    #     text : dict, optional
    #         The dictionary to print, if None, the keys are displayed, by default None
    #     fontsize : int, optional
    #         Font size, by default 12
    #     """

    #     if not self.formartist:
    #         return

    #     self.formartist.plotter.fontsize = fontsize
    #     self.formartist.draw_vertexlabels(text=text)

    def highlight_vertices(self, vertices, show_forcepolygon=False):
        """Highlight a vertex in the form diagram and in the force diagram (if they have been previously added to the plotter)

        Parameters
        ----------
        vertices : [int]
            Key of the vertices to highlight
        show_forcepolygon : bool, optional
            If force polygons should be also displayed in the force diagram, by default False
            Note: this requires that the force diagram is already added to the scene.

        Returns
        -------
        None
            The Plotter is updated in place.
        """

        points = [self.form.vertex_coordinates(vertex) for vertex in vertices]
        edges = [self.form.vertex_edges(vertex) for vertex in vertices]
        base_thick = self.settings['size.edge.base_thickness']

        for i, point in enumerate(points):
            pt = Point(point[0], point[1], point[2])
            self.app.add(pt, facecolor=(0.8, 0.8, 0.8), size=self.settings['size.vertex'])

            for edge in edges[i]:
                line = Line(self.form.vertex_coordinates(edge[0]), self.form.vertex_coordinates(edge[1]))
                self.app.add(line, color=(0, 0, 0), linewidth=base_thick * 3, draw_as_segment=True)

        if show_forcepolygon:
            if self.force:
                vertex_index = self.form.vertex_index()
                for vertex in vertices:
                    i = vertex_index[vertex]
                    force_faces = list(self.force.faces())
                    face = self.force.face_coordinates(force_faces[i])
                    polygon = Polygon(face)
                    S = Scale.from_factors(3 * [self.scale_forcediagram])
                    T = Translation.from_vector(self.translation_forcediagram)
                    polygon = polygon.transformed(S)
                    polygon = polygon.transformed(T)
                    self.app.add(polygon, linewidth=base_thick * 3, facecolor=(0.8, 0.8, 0.8))

    def draw_form_sym(self, print_sym=False):
        """Draw the form diagram symmetry edges with the respective colors.

        Parameters
        ----------
        print_sym : bool, optional
            If text should be added to the symmetry edges, by default True

        Returns
        -------
        None
            The plotter is updated in place.
        """

        texts = {}
        colors = {}
        edges = []

        base_thick = self.settings['size.edge.base_thickness'] * 2

        i_sym_max = 0
        for edge in self.form.edges_where({'_is_edge': True}):
            edges.append(edge)
            i_sym = self.form.edge_attribute(edge, 'sym_key')
            if i_sym is None:
                raise NameError('Check if symmetry is applied to to the problem formulation.')
            if i_sym > i_sym_max:
                i_sym_max = i_sym

        colormap = plt.cm.get_cmap('hsv')  # gist_ncar nipy_spectral, Set1, Paired coolwarm
        linspace = [1.0/(i_sym_max) * i for i in range(i_sym_max + 1)]
        colormaps = [rgb_to_hex(colormap(i)[:3]) for i in linspace]

        for edge in edges:
            i_sym = self.form.edge_attribute(edge, 'sym_key')
            colors[edge] = colormaps[i_sym]
            if print_sym:
                texts[edge] = str(i_sym)

        self.formartist = self.app.add(
            self.form,
            edges=edges,
            edgewidth=base_thick,
            # edge_text=texts,  # not working for COMPAS 1.15
            edgecolor=colors,
            show_vertices=False,
            show_faces=False
        )
        if print_sym:
            self.formartist.draw_edgelabels(text=texts)

    def draw_force(self, scale=None, show_edges=True, show_vertices=False, show_faces=False, show_independents=False):
        """Draw the force diagram associated with the form diagram.

        Parameters
        ----------
        scale : float, optional
            If a scale should be defined. If None, a convenient scale is compuuted, by default None
        show_edges : bool, optional
            Whether or not edges are shown in the force diagram, by default True
        show_vertices : bool, optional
            Whether or not vertices are shown in the force diagram, by default False
        show_faces : bool, optional
            Whether or not faces are shown in the force diagram, by default False
        show_independentrs : bool, optional
            Whether or not independent edges are highlighted, by default False

        Returns
        -------
        None
            The plotter is updated in place.
        """
        if not self.force:
            from compas_tno.algorithms import reciprocal_from_form
            self.force = reciprocal_from_form(self.form)

        force = self.force

        base_width = self.settings['size.edge.base_thickness']
        width = {}

        edgecolor = {}
        for edge in force.edges():
            edgecolor[edge] = Color.black()
            width[edge] = base_width
        if show_independents:
            for edge in force.edges_where_dual({'is_ind': True}):
                edgecolor[edge] = self.settings['color.edges.independent']
                width[edge] = base_width * 3

        force_bbox = self.force.bounding_box_xy()
        xmin_force, xmax_force = force_bbox[0][0], force_bbox[1][0]
        ymin_force, ymax_force = force_bbox[0][1], force_bbox[2][1]
        dx_force, dy_force = (xmax_force - xmin_force), (ymax_force - ymin_force)

        if self.form:

            form_bbox = self.form.bounding_box_xy()
            xmin_form, xmax_form = form_bbox[0][0], form_bbox[1][0]
            ymin_form, ymax_form = form_bbox[0][1], form_bbox[2][1]
            dx_form, dy_form = (xmax_form - xmin_form), (ymax_form - ymin_form)

            scale = scale or min(dy_form/dy_force, dx_form/dx_force)
            print('Diagram scale factor:', scale)

            tx = dx_form * 1.2 + (xmax_form - xmin_force) * scale
            ty = (ymin_form - ymin_force) * scale

            frame = Frame([xmin_form, ymin_form, 0.0], (1, 0, 0), (0, 1, 0))

            translation = [tx, ty, 0]

            self.scale_forcediagram = scale
            self.translation_forcediagram = translation

            S = Scale.from_factors(3 * [scale], frame=frame)
            T = Translation.from_vector(translation)

            force = force.transformed(S)
            force = force.transformed(T)

        self.app.add(force,
                     show_edges=show_edges,
                     show_vertices=show_vertices,
                     show_faces=show_faces,
                     edgecolor=edgecolor,
                     edgewidth=width,
                     )

    def draw_edgelabels(self, text=None):
        """Draw Labels to the vertices

        Parameters
        ----------
        text : dict, optional
            The dictionary to print, if None, the keys are displayed, by default None
        """

        if not self.formartist:
            return

        self.formartist.draw_edgelabels(text=text)

    def draw_arch_lines(self, H=1.00, L=2.0, x0=0.0, thk=0.20, total_nodes=50, stereotomy=False, close_bottom=True):
        """Helper to draw the lines of intrados and extrados of an arch for given parameters

        Parameters
        ----------
        H : float, optional
            Height of the arch, by default 1.00
        L : float, optional
            Span of the arch, by default 2.0
        x0 : float, optional
            Starting coordinate of the arch , by default 0.0
        thk : float, optional
            Thickness of the arch, by default 0.20
        total_nodes : int, optional
            Density of the shape, equals to the number of blocks, by default 50
        stereotomy : bool, optional
            Whether or not interfaces of the stereotomy should be drawn, by default False
        close_bottom : bool, optional
            Whether or not the last interfaces of the arch should be drawn, by default True

        Returns
        -------
        None
            Lines are added to the plotter
        """

        lines_intrados = []
        lines_extrados = []

        radius = H / 2 + (L**2 / (8 * H))
        ri = radius - thk/2
        re = radius + thk/2
        spr = math.atan2((L/2), (radius - H))
        tot_angle = 2*spr
        angle_init = (math.pi - tot_angle)/2
        an = tot_angle / (total_nodes)
        zc = radius - H
        xc = L/2 + x0
        i = 0

        for i in range(total_nodes):
            angle_i = angle_init + i * an
            angle_f = angle_init + (i + 1) * an

            xii = xc - ri * math.cos(angle_i)
            xif = xc - ri * math.cos(angle_f)
            zii = ri * math.sin(angle_i) - zc
            zif = ri * math.sin(angle_f) - zc

            xei = xc - re * math.cos(angle_i)
            xef = xc - re * math.cos(angle_f)
            zei = re * math.sin(angle_i) - zc
            zef = re * math.sin(angle_f) - zc

            line_i = Line([xii, zii, 0.0], [xif, zif, 0.0])
            lines_intrados.append(line_i)

            line_e = Line([xei, zei, 0.0], [xef, zef, 0.0])
            lines_extrados.append(line_e)

        interfaces = []

        if close_bottom:
            interfaces = [0, total_nodes]
        if stereotomy:
            interfaces = list(range(total_nodes + 1))

        for i in interfaces:
            angle = angle_init + i * an

            xi = xc - ri * math.cos(angle)
            xf = xc - re * math.cos(angle)
            zi = ri * math.sin(angle) - zc
            zf = re * math.sin(angle) - zc

            line = Line([xi, zi, 0.0], [xf, zf, 0.0])
            lines_intrados.append(line)

        # update this when collections are available

        for le in lines_extrados:
            self.app.add(le, draw_as_segment=True, opacity=self.settings['shape.opacity'], color=self.settings['color.edges.shape'])
        for li in lines_intrados:
            self.app.add(li, draw_as_segment=True, opacity=self.settings['shape.opacity'], color=self.settings['color.edges.shape'])

        return

    def make_gif(self, images_file, filepath, interval=10):
        """Make a gif from a list of image files

        Parameters
        ----------
        images_file : str
            List with the files to make the image. Note that list should be ordered
        filepath : str
            Pathh to save the .gif file
        interval : int, optional
            Interval between images, by default 10

        Returns
        -------
        None
            GIF is saved in the address provided
        """

        images = []
        for path in images_file:
            images.append(Image.open(path))

        images[0].save(filepath, save_all=True, append_images=images[1:], optimize=False, duration=interval * 10, loop=0)



from compas_tno.algorithms import compute_reactions
from compas_tno.algorithms import xyz_from_q
from compas_tno.algorithms import q_from_variables
from compas_tno.shapes import Shape
from compas_tno.shapes.meshdos import MeshDos
from compas_tno.utilities import apply_envelope_from_shape
from compas_tno.problems import save_geometry_at_iterations
from compas_tno.utilities.envelopes import apply_bounds_reactions


def post_process_general(analysis):
    """Post processing of the optimisation.

    Parameters
    ----------
    analysis : Analysis
        The Analysis object

    Returns
    -------
    analysis : Analysis
        The Analysis object updated
    """

    form = analysis.form
    optimiser = analysis.optimiser
    shape = analysis.shape

    M = optimiser.M
    summary = optimiser.settings.get('summary', False)
    printout = optimiser.settings.get('printout', True)
    thickness_type = optimiser.settings.get('thickness_type', 'constant')
    features = optimiser.settings.get('features', [])
    save_iterations = optimiser.settings.get('save_iterations', False)
    show_force_diagram = optimiser.settings.get('save_force_diagram', True)

    fconstr = optimiser.fconstr
    # args = optimiser.args
    xopt = optimiser.xopt
    fopt = optimiser.fopt
    message = optimiser.message
    thk = M.thk

    i_uv = form.index_uv()
    # i_k = form.index_key()

    check = M.k
    qid = xopt[:M.k].reshape(-1, 1)

    if 'xyb' in M.variables:
        xyb = xopt[check:check + 2*M.nb]
        check = check + 2*M.nb
        M.X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
    if 'zb' in M.variables:
        zb = xopt[check: check + M.nb]
        check = check + M.nb
        M.X[M.fixed, [2]] = zb.flatten()
    if 't' in M.variables:
        thk = xopt[-1]
    if 'n' in M.variables:
        n = xopt[-1]
    if 'lambdh' in M.variables:
        lambdh = xopt[check: check + 1]
        M.P[:, [0]] = lambdh * M.px0
        M.P[:, [1]] = lambdh * M.py0
        M.d = lambdh * M.d0
    if 'lambdv' in M.variables:
        lambdv = xopt[check: check + 1]
        M.P[:, [2]] = lambdv * M.pzv + M.pz0
    if 'tub' in M.variables:
        tub = xopt[check:check + M.n]
        check = check + M.n
    if 'tlb' in M.variables:
        tlb = xopt[check:check + M.n]
        check = check + M.n
    if 'tub_reac' in M.variables:
        tub_reac = xopt[check: check + 2*M.nb]
        check = check + 2*M.nb

    M.q = q_from_variables(qid, M.B, M.d)

    # ADD 'lambdv' to post-process

    # if 's' in variables:
    #     s = xopt[-1]

    g_final = fconstr(xopt, M)
    M.X[M.free] = xyz_from_q(M.q, M.P[M.free], M.X[M.fixed], M.Ci, M.Cit, M.Cb)

    # if printout:
    #     print('post-processing min, max z:', min(M.X[:, 2].flatten()), max(M.X[:, 2].flatten()))

    i = 0
    for key in form.vertices():
        form.vertex_attribute(key, 'x', M.X[i, 0])
        form.vertex_attribute(key, 'y', M.X[i, 1])
        form.vertex_attribute(key, 'z', M.X[i, 2])
        form.vertex_attribute(key, 'px', M.P[i, 0])
        form.vertex_attribute(key, 'py', M.P[i, 1])
        form.vertex_attribute(key, 'pz', M.P[i, 2])
        i = i + 1

    for c, qi in enumerate(list(M.q.ravel())):
        u, v = i_uv[c]
        li = form.edge_length(u, v)
        form.edge_attribute((u, v), 'q', float(qi))
        form.edge_attribute((u, v), 'f', float(qi*li))

    form.attributes['loadpath'] = form.loadpath()
    compute_reactions(form)

    if 't' in M.variables:
        if shape.datashape['type'] == 'general':
            if thickness_type == 'constant':
                form.attributes['thk'] = thk
                shape.datashape['thk'] = thk
                shape.intrados = shape.middle.offset_mesh(n=thk/2, direction='down')
                shape.extrados = shape.middle.offset_mesh(n=thk/2, direction='up')
                apply_envelope_from_shape(form, shape)
            elif thickness_type == 'variable':
                t0 = shape.datashape['thk']
                thk = t0 * thk  # Consider that the thk for general shapes is a percentage of the thickness
                form.attributes['thk'] = thk
                shape.datashape['thk'] = thk
                if printout:
                    print('Optimum Value corresponds to a thickness of:', thk)
                shape.extrados, shape.intrados = shape.middle.offset_up_and_down(n=fopt)
                apply_envelope_from_shape(form, shape)
            elif thickness_type == 'intrados':
                form.attributes['thk'] = thk
                shape.datashape['thk'] = thk
                shape.middle = shape.intrados.offset_mesh(n=thk/2, direction='up')
                shape.extrados = shape.intrados.offset_mesh(n=thk, direction='up')
                form.envelope_from_shape(shape)
        else:
            form.attributes['thk'] = thk
            shape.datashape['thk'] = thk
            shape = Shape.from_library(shape.datashape)
            apply_envelope_from_shape(form, shape)  # Check if this is ok for adapted pattern
            apply_bounds_reactions(form, shape)
            i = 0
            for key in form.vertices():  # this resolve the problem due to the adapted pattern
                form.vertex_attribute(key, 'ub', float(M.ub[i]))
                form.vertex_attribute(key, 'lb', float(M.lb[i]))
                i += 1

    if 'update-envelope' in features:
        form.attributes['thk'] = thk
        shape.datashape['thk'] = thk
        shape = Shape.from_library(shape.datashape)
        apply_envelope_from_shape(form, shape)

    # if 's' in M.variables:
    #     s = -1 * fopt
    #     for key in form.vertices():
    #         ub = form.vertex_attribute(key, 'ub')
    #         lb = form.vertex_attribute(key, 'lb')
    #         form.vertex_attribute(key, 'ub', ub - s * (ub - lb))
    #         form.vertex_attribute(key, 'lb', lb + s * (ub - lb))

    if 'n' in M.variables:
        print('Value of N:', n)
        n = -1 * fopt
        shape.intrados: MeshDos = shape.intrados.offset_mesh(n=n, direction='up')
        shape.extrados: MeshDos = shape.extrados.offset_mesh(n=n, direction='down')
        apply_envelope_from_shape(form, shape)

    if 'tub' in M.variables:
        for i, key in enumerate(form.vertices()):
            zub = form.vertex_attribute(key, 'ub')
            form.vertex_attribute(key, 'tub', tub[i])
            form.vertex_attribute(key, 'ub', zub + tub[i])

    if 'tlb' in M.variables:
        for i, key in enumerate(form.vertices()):
            zub = form.vertex_attribute(key, 'lb')
            form.vertex_attribute(key, 'tlb', tlb[i])
            form.vertex_attribute(key, 'lb', zub - tlb[i])

    if 'tub_reac' in M.variables:
        for i, key in enumerate(form.vertices_where({'is_fixed': True})):
            print(i, key, tub_reac)
            form.vertex_attribute(key, 'tub_reac', [tub_reac[i], tub_reac[i + M.nb]])

    if 'lambdv' in M.variables:
        added_load = lambdv * M.pzv
        for i, key in enumerate(form.vertices()):
            p_added = added_load[i]
            form.vertex_attribute(key, 'pzext', p_added)

    if 'lambdh' in M.variables:
        form.attributes['lambdh'] = lambdh  # can be improved. It currently takes fopt, but if loads at start are 0.1*SWT, the obj is lambd/0.1

    analysis.form = form
    analysis.optimiser = optimiser
    analysis.shape = shape

    if save_iterations:
        file_Xform, file_Xforce = save_geometry_at_iterations(form, optimiser, force=show_force_diagram)
        analysis.optimiser.Xform = file_Xform
        analysis.optimiser.Xforce = file_Xforce

    if printout or summary:
        print('\n' + '-' * 50)
        print('Solution  :', message)
        try:
            print('q range : {0:.3f} : {1:.3f}'.format(min(M.q), max(M.q)))
        except BaseException:
            print('q range : {0:.3f} : {1:.3f}'.format(min(M.q.flatten()), max(M.q.flatten())))
        print('zb range  : {0:.3f} : {1:.3f}'.format(min(M.X[M.fixed, [2]]), max(M.X[M.fixed, [2]])))
        print('constr    : {0:.3f} : {1:.3f}'.format(min(g_final), max(g_final)))
        print('fopt      : {0:.3f}'.format(fopt))
        print('-' * 50 + '\n')

    return analysis



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino

from compas_tno.rhino import SettingsForm
from compas_tno.rhino import FormObject
from compas_tno.rhino import ShapeObject
from compas_tno.rhino import OptimiserObject
from compas_tno.rhino import ForceObject


__commandname__ = "TNO_settings"


def RunCommand(is_interactive):

    if 'TNO' not in sc.sticky:
        compas_rhino.display_message('TNO has not been initialised yet.')
        return

    scene = sc.sticky['TNO']['scene']
    if not scene:
        return

    SettingsForm.from_scene(scene, object_types=[FormObject, ForceObject, ShapeObject, OptimiserObject], global_settings=['TNO'])


# ==============================================================================
# Main
# ==============================================================================
if __name__ == '__main__':

    RunCommand(True)



"""
********************************************************************************
compas_tno.problems
********************************************************************************

.. currentmodule:: compas_tno.problems

Classes
=======

.. autosummary::
    :toctree: generated/

    Problem

Initialisation
==============

.. autosummary::
    :toctree: generated/

    initialize_fdm
    initialize_tna
    initialize_loadpath
    initialise_form
    initialise_problem_general
    adapt_problem_to_fixed_diagram
    adapt_problem_to_sym_diagram
    adapt_problem_to_sym_and_fixed_diagram
    apply_sym_to_form

Set up
======

.. autosummary::
    :toctree: generated/

    set_up_general_optimisation
    set_up_convex_optimisation

Starting Points
===============

.. autosummary::
    :toctree: generated/

    initialize_loadpath
    initialize_tna

Objectives
==========

.. autosummary::
    :toctree: generated/

    objective_selector
    f_min_thrust
    f_max_thrust
    f_bestfit
    f_horprojection
    f_loadpath_general
    f_complementary_energy
    f_complementary_energy_nonlinear
    f_max_section
    f_constant
    f_reduce_thk
    f_tight_crosssection

Constraints
===========

.. autosummary::
    :toctree: generated/

    constr_wrapper

Gradients
=========

.. autosummary::
    :toctree: generated/

    d_fobj
    compute_dQ
    gradient_feasibility
    gradient_reduce_thk
    gradient_tight_crosssection
    gradient_fmin
    gradient_fmax
    gradient_bestfit
    gradient_horprojection
    gradient_complementary_energy
    gradient_complementary_energy_nonlinear
    gradient_loadpath
    gradient_max_section

Jacobian
========

.. autosummary::
    :toctree: generated/

    d_fconstr
    sensitivities_wrapper

Proxy
=====

.. autosummary::
    :toctree: generated/

    initialize_loadpath_proxy
    run_NLP_proxy

Callbacks
=========

.. autosummary::
    :toctree: generated/

    callback_save_json
    callback_create_json
    save_geometry_at_iterations

"""

from .bounds_update import (
    ub_lb_update,
    dub_dlb_update,
    b_update,
    db_update
)

from .constraints import (
    constr_wrapper
)

from .derivatives import (
    d_fobj,
    compute_dQ,
    gradient_feasibility,
    gradient_reduce_thk,
    gradient_tight_crosssection,
    gradient_fmin,
    gradient_fmax,
    gradient_bestfit,
    gradient_horprojection,
    gradient_complementary_energy,
    gradient_complementary_energy_nonlinear,
    gradient_loadpath,
    gradient_max_section
)

from .jacobian import (
    d_fconstr,
    sensitivities_wrapper
)

from .objectives import (
    objective_selector,
    f_min_thrust,
    f_max_thrust,
    f_bestfit,
    f_horprojection,
    f_loadpath_general,
    f_complementary_energy,
    f_complementary_energy_nonlinear,
    f_max_section,
    f_constant,
    f_reduce_thk,
    f_tight_crosssection,
)

from .problems import (
    Problem,
    initialise_form,
    initialise_problem_general,
    adapt_problem_to_fixed_diagram,
    adapt_problem_to_sym_diagram,
    adapt_problem_to_sym_and_fixed_diagram,
    apply_sym_to_form,
)

from .callbacks import (
    callback_save_json,
    callback_create_json,
    save_geometry_at_iterations
)

from .initialize import (
    initialize_loadpath,
    initialize_tna,
    initialize_fdm
)

from .setup import (
    set_up_general_optimisation,
    set_up_convex_optimisation,
)

from .proxy import (
    initialize_loadpath_proxy,
    run_NLP_proxy
)


__all__ = [

    'ub_lb_update',
    'dub_dlb_update',
    'b_update',
    'db_update',

    'constr_wrapper',

    'd_fobj',
    'compute_dQ',
    'gradient_feasibility',
    'gradient_reduce_thk',
    'gradient_tight_crosssection',
    'gradient_fmin',
    'gradient_fmax',
    'gradient_bestfit',
    'gradient_horprojection',
    'gradient_complementary_energy',
    'gradient_complementary_energy_nonlinear',
    'gradient_loadpath',
    'gradient_max_section',

    'd_fconstr',
    'sensitivities_wrapper',

    'objective_selector',
    'f_min_thrust',
    'f_max_thrust',
    'f_bestfit',
    'f_horprojection',
    'f_loadpath_general',
    'f_complementary_energy',
    'f_complementary_energy_nonlinear',
    'f_max_section',
    'f_constant',
    'f_reduce_thk',
    'f_tight_crosssection',

    'Problem',
    'initialise_form',
    'initialise_problem_general',
    'adapt_problem_to_fixed_diagram',
    'adapt_problem_to_sym_diagram',
    'adapt_problem_to_sym_and_fixed_diagram',
    'apply_sym_to_form',

    'callback_save_json',
    'callback_create_json',
    'save_geometry_at_iterations',

    'initialize_loadpath',
    'initialize_tna',
    'initialize_fdm',

    'set_up_general_optimisation',
    'set_up_convex_optimisation',

    'initialize_loadpath_proxy',
    'run_NLP_proxy'

]



import time
from compas_tno.shapes import Shape
from compas_tno.diagrams import FormDiagram
import compas_tno
import os
import math


def limit_analysis_GSF(analysis, thk, thk_reduction, thk_refined=None, limit_equal=0.01, printout=True, save_forms=False):
    """Routine to compute the succesive max/min optimisation optimisation.

    Parameters
    ----------
    analysis : :class:`~compas_tno.analysis.Analysis`
        The Analysis object
    thk : float
        The initial thickness of the structure
    thk_reduction : float
        Step of reduction of the thickness
    thk_refined : float, optional
        If require a smaller step close to convergence, by default None
    limit_equal : float, optional
        Tolerance for the stopping criteria where the min/max solutions are considered equal, by default 0.01
    printout : bool, optional
        Whether or not printing in the screen the solutions obtained, by default True
    save_forms : bool, optional
        Whether or not solutions should be saved, by default False

    Returns
    -------
    [list, list], [list, list]
        Lists with the thicknesses obtained in min/max computation and the soulutions normalised of min/max thrust.
    """

    solutions_min = []  # empty lists to keep track of the solutions for min thrust
    solutions_max = []  # empty lists to keep track of the solutions for max thrust
    thicknesses_min = []
    thicknesses_max = []
    t0 = thk
    form0 = analysis.form.copy()
    thk_reduction0 = thk_reduction
    data_diagram = analysis.form.parameters
    data_shape = analysis.shape.datashape
    ro = analysis.shape.ro
    last_min = 0
    last_max = 100
    last_thk_min = t0
    objectives = ['min', 'max']
    exitflag = 0
    address = None

    print('Limit Analysis - GSF: For ({0}) with diagram ({1})'.format(data_shape['type'], data_diagram['type']))

    while (last_max - last_min) > limit_equal:

        if exitflag == 0:
            pass
        else:
            objectives = objectives[::-1]
            print('Warning: Did not Achieve precision required. Stopping Anyway.\n')
            break

        for analysis.optimiser.settings['objective'] in objectives:

            analysis.form = form0
            exitflag = 0
            thk = t0
            thk_reduction = thk_reduction0
            count = 0

            print('\n----- Starting the [', analysis.optimiser.settings['objective'], '] problem for intial thk:', thk)
            print('THK  |   Solved  |   Opt.Val |   Opt/W   |   THK red.  |   Setup time  |   Run time')

            while exitflag == 0 and thk > 0:

                # analysis.form = form0  # added for general case
                data_shape['thk'] = thk
                time0 = time.time()
                analysis.shape = Shape.from_library(data_shape)
                analysis.shape.ro = ro
                swt = analysis.shape.compute_selfweight()
                analysis.apply_selfweight()
                analysis.apply_envelope()
                analysis.apply_reaction_bounds()
                analysis.set_up_optimiser()
                setup_time = time.time() - time0
                analysis.run()
                run_time = time.time() - time0 - setup_time
                exitflag = analysis.optimiser.exitflag  # get info if optimisation was succeded ot not
                fopt = analysis.optimiser.fopt  # objective function optimum value
                fopt_over_weight = fopt/swt  # divide by selfweight

                if exitflag == 0:
                    if analysis.optimiser.settings['objective'] == 'min':
                        solutions_min.append(fopt_over_weight)
                        thicknesses_min.append(thk)
                        last_min = fopt_over_weight
                        last_thk_min = thk
                    else:
                        solutions_max.append(fopt_over_weight)
                        thicknesses_max.append(thk)
                        last_max = abs(fopt_over_weight)
                    if save_forms:
                        address = save_forms + '_' + analysis.optimiser.settings['objective'] + '_thk_' + str(100*thk) + '.json'
                    else:
                        address = os.path.join(compas_tno.get('/temp/'), 'form_' + analysis.optimiser.settings['objective'] + '.json')
                    analysis.form.to_json(address)
                    print('{0}  |   True  |   {1:.1f} |   {2:.6f} |   {3}   | {4:.2f}s  |   {5:.2f}s'.format(thk, fopt, fopt_over_weight, thk_reduction, setup_time, run_time))
                    thk = round(thk - thk_reduction, 5)
                else:
                    if not address:
                        if analysis.optimiser.settings['objective'] == objectives[0]:
                            print('Failed in Initial Thickness and objective ({0})'.format(analysis.optimiser.settings['objective']))
                            objectives = objectives[::-1]
                    elif analysis.optimiser.settings['objective'] == 'max' and last_thk_min <= thk:
                        print('{0}  |   False  |   XXXX |   XXXX |   Load next [min]  | {1:.2f}s   |  {2:.2f}s'.format(thk, setup_time, run_time))
                        # print('Try loading next [min] optimisation - 1')
                        next_min = None
                        for i in range(len(thicknesses_min)):
                            if thicknesses_min[i] < thk or (thicknesses_min[i] == thk and count == 0):
                                next_min = thicknesses_min[i]
                                break
                        if next_min:
                            address = save_forms + '_' + 'min' + '_thk_' + str(100*next_min) + '.json'
                            analysis.form = FormDiagram.from_json(address)
                            thk = next_min
                            exitflag = 0
                        else:
                            print('Tried loading all [min] optimisation')
                            print('---------- End of process -----------', '\n')
                    elif thk_reduction > thk_refined:  # or (analysis.optimiser.settings['objective'] == 'max' and last_thk_min < last_thk_max and thk_reduction > thk_refined/2):
                        analysis.form = FormDiagram.from_json(address)  # reload last solved
                        thk_ = thk
                        thk = round(thk + thk_reduction, 5)
                        thk_reduction = thk_reduction / 2
                        print('{0}  |   False  |   XXXX |   XXXX |   {1}    | {2:.2f}s   |  {3:.2f}s'.format(thk_, thk_reduction, setup_time, run_time))
                        thk = round(thk - thk_reduction, 5)
                        exitflag = 0
                    else:
                        print('---------- End of process -----------', '\n')
                count += 1

    if printout:
        print('------------------ SUMMARY ------------------ ')
        print('ANALYSIS FOR THICKNESS t0:', t0)
        print('thicknesses min ({0}):'.format(len(thicknesses_min)))
        print(thicknesses_min)
        print('thicknesses max ({0}):'.format(len(thicknesses_max)))
        print(thicknesses_max)
        print('min solutions ({0}):'.format(len(solutions_min)))
        print(solutions_min)
        print('max solutions ({0}):'.format(len(solutions_max)))
        print(solutions_max)

    return [thicknesses_min, thicknesses_max],  [solutions_min, solutions_max]


def thk_minmax_GSF(analysis, thk_max, thk_step=0.05, printout=True, save_forms=None, skip_minthk=False):
    """Routine to compute the succesive max/min optimisations starting from the minimum thickness.

    Parameters
    ----------
    analysis : :class:`~compas_tno.analysis.Analysis`
        The Analysis object
    thk_max : float
        Maximum thickness, in which the algorithm should stop
    thk_step : float, optional
        The increasing steps of the algorithm, by default 0.05
    printout : bool, optional
        If prints should be displayed in the screen, by default True
    save_forms : bool, optional
        If forms should be saved as JSON, by default None
    skip_minthk : bool, optional
        If the minimum thickness computation can be skiped (it assumes the minimum thickness as it is), by default False

    Returns
    -------
    [list, list], [list, list]
        Lists with the thicknesses obtained in min/max computation and the soulutions normalised of min/max thrust.
    """

    solutions_min = []  # empty lists to keep track of the solutions for min thrust
    solutions_max = []  # empty lists to keep track of the solutions for max thrust
    thicknesses_min = []
    thicknesses_max = []
    objectives = ['min', 'max']
    ro = analysis.shape.ro
    data_shape = analysis.shape.datashape

    # Find extreme (min thickness) solution:

    print('\n----- Starting the min thk optimisation for starting thk: {0:.4f}'.format(analysis.shape.data['thk']))

    if not skip_minthk:
        time0 = time.time()
        analysis.optimiser.settings['variables'] = ['ind', 'zb', 't']
        analysis.optimiser.settings['objective'] = 't'
        analysis.apply_selfweight()
        analysis.apply_envelope()
        analysis.apply_reaction_bounds()
        analysis.set_up_optimiser()
        setup_time = time.time() - time0
        analysis.run()
        run_time = time.time() - time0 - setup_time
        exitflag = analysis.optimiser.exitflag
    else:
        exitflag = 0
        setup_time = 0.0
        run_time = 0.0

    if exitflag == 0:
        thk_min = analysis.form.attributes['thk']
        # compute_reactions(analysis.form)
        swt = analysis.form.lumped_swt()
        T = analysis.form.thrust()
        T_over_swt = T/swt
        thk_increase = int(thk_step*100)*math.ceil(thk_min*100.0/int(thk_step*100))/100.0 - thk_min
        print('Min THK  |   Solved  |   Thrust |   T/W   |  THK incr. |  Setup time  |   Run time')
        print('{0:.5f}  |   True  |   {1:.1f} |   {2:.6f} |   {3:.2f}   | {4:.2f}s  |   {5:.2f}s'.format(thk_min, T, T_over_swt, thk_increase, setup_time, run_time))

        # STORE

        address0 = os.path.join(compas_tno.get('/temp/'), 'form0.json')
        analysis.form.to_json(address0)

        solutions_min.append(T_over_swt)
        solutions_max.append(-1 * T_over_swt)
        thicknesses_min.append(thk_min)
        thicknesses_max.append(thk_min)

        # SAVE
        if save_forms:
            address_min = save_forms + '_' + 'min' + '_thk_' + str(100*thk_min) + '.json'
            address_max = save_forms + '_' + 'max' + '_thk_' + str(100*thk_min) + '.json'
            analysis.form.to_json(address_min)
            analysis.form.to_json(address_max)

        thk = round(thk_min + thk_increase, 5)
        thk_increase = thk_step
        thk0 = thk

        if thk_min == thk_max:
            print('Warning: Minimum THK Optimisation found optimum at start. Try rescalling the problem.')
            return
    else:
        print('Error: Minimum THK Optimisation did not find a solution: Try scalling the optimisation, or starting in a different thickness value')
        return

    # Start inverse loop for min/max:

    analysis.optimiser.settings['variables'] = ['ind', 'zb']
    for analysis.optimiser.settings['objective'] in objectives:

        analysis.form = FormDiagram.from_json(address0)

        exitflag = 0
        thk = thk0
        count = 0
        first_fail = True

        print('\n----- Starting the inverse [', analysis.optimiser.settings['objective'], '] problem for minimum thk:', thk)
        print('THK  |   Solved  |   Opt.Val |   Opt/W   |   THK incr.  |   Setup time  |   Run time')

        while exitflag == 0 and thk <= thk_max:

            data_shape['thk'] = thk

            time0 = time.time()
            analysis.shape = Shape.from_library(data_shape)
            analysis.shape.ro = ro

            swt = analysis.shape.compute_selfweight()

            lumped_swt = analysis.form.lumped_swt()

            analysis.form.scale_form(swt/lumped_swt)
            lumped_swt = analysis.form.lumped_swt()

            analysis.apply_envelope()
            analysis.apply_reaction_bounds()
            analysis.set_up_optimiser()
            setup_time = time.time() - time0
            analysis.run()
            run_time = time.time() - time0 - setup_time
            exitflag = analysis.optimiser.exitflag  # get info if optimisation was succeded ot not
            fopt = analysis.optimiser.fopt  # objective function optimum value
            fopt_over_weight = fopt/lumped_swt  # divide by selfweight

            if exitflag == 0:
                if analysis.optimiser.settings['objective'] == 'min':
                    solutions_min.append(fopt_over_weight)
                    thicknesses_min.append(thk)
                else:
                    solutions_max.append(fopt_over_weight)
                    thicknesses_max.append(thk)
                if save_forms:
                    address = save_forms + '_' + analysis.optimiser.settings['objective'] + '_thk_' + str(100*thk) + '.json'
                else:
                    address = os.path.join(compas_tno.get('/temp/'), 'form_' + analysis.optimiser.settings['objective'] + '.json')
                analysis.form.to_json(address)
                print('{0:.5f}  |   True  |   {1:.1f} |   {2:.6f} |   {3}   | {4:.2f}s  |   {5:.2f}s'.format(thk, fopt, fopt_over_weight, thk_increase, setup_time, run_time))
                thk = thk + thk_increase
                first_fail = True
            else:
                print('Failed Optimisation for [{0}] with thk: {1}'.format(analysis.optimiser.settings['objective'], thk))
                other_objective = 'min' if analysis.optimiser.settings['objective'] == 'max' else 'max'
                if thk < thk_max:
                    if not first_fail:
                        thk = thk + thk_increase
                        first_fail = True
                        print('Second Fail with [{0}] opt, will load [{1}] opt in incr. thk: {2}'.format(analysis.optimiser.settings['objective'], other_objective, thk))
                    else:
                        print('First Fail with [{0}] opt, will load [{1}] opt in thk: {2}'.format(analysis.optimiser.settings['objective'], other_objective, thk))
                        first_fail = False
                    address_other_obj = save_forms + '_' + other_objective + '_thk_' + str(100*thk) + '.json'
                    analysis.form = FormDiagram.from_json(address_other_obj)
                    exitflag = 0
            count += 1
        print('---------- End of process -----------', '\n')

    thicknesses_min.reverse()
    thicknesses_max.reverse()
    solutions_min.reverse()
    solutions_max.reverse()

    if printout:
        print('------------------ SUMMARY ------------------ ')
        print('ANALYSIS FOUND MIN THK:', thk_min)
        print('thicknesses min ({0}):'.format(len(thicknesses_min)))
        print(thicknesses_min)
        print('thicknesses max ({0}):'.format(len(thicknesses_max)))
        print(thicknesses_max)
        print('min solutions ({0}):'.format(len(solutions_min)))
        print(solutions_min)
        print('max solutions ({0}):'.format(len(solutions_max)))
        print(solutions_max)

    return [thicknesses_min, thicknesses_max],  [solutions_min, solutions_max]


def max_n_minmax_GSF(analysis, n_step=0.01, printout=True, save_forms=False):
    """Routine to compute max/min thrust optimisation problems for non-analytic surfaces, considering the nomal vectors 'n'.

    Parameters
    ----------
    analysis : :class:`~compas_tno.analysis.Analysis`
        The Analysis object
    n_step : float, optional
        The step to decrease the normal vector, by default 0.01
    printout : bool, optional
        If prints are added to the screen, by default True
    save_forms : bool, optional
        If forms are saved as JSON, by default False

    Returns
    -------
    [list, list], [list, list]
        Lists with the thicknesses obtained in min/max computation and the soulutions normalised of min/max thrust.
    """

    solutions_min = []  # empty lists to keep track of the solutions for min thrust
    solutions_max = []  # empty lists to keep track of the solutions for max thrust
    thicknesses_min = []
    thicknesses_max = []
    objectives = ['min', 'max']
    ro = analysis.shape.ro
    data_shape = analysis.shape.datashape
    thk0 = data_shape['thk']
    t = data_shape['t']
    initial_intrados = analysis.shape.intrados.copy()
    initial_extrados = analysis.shape.extrados.copy()
    middle = analysis.shape.middle.copy()

    # Find extreme (min thickness) solution:

    print('\n----- Starting the min thk optimisation for approx. thk: {0:.4f}'.format(thk0))

    time0 = time.time()
    analysis.optimiser.settings['variables'] = ['ind', 'zb', 'n']
    analysis.optimiser.settings['objective'] = 'n'
    analysis.apply_selfweight()
    analysis.apply_envelope()
    analysis.apply_reaction_bounds()
    analysis.set_up_optimiser()
    setup_time = time.time() - time0
    analysis.run()
    run_time = time.time() - time0 - setup_time

    exitflag = analysis.optimiser.exitflag

    if exitflag == 0:
        n = - 1 * analysis.optimiser.fopt
        thk = thk0 - 2 * n
        thk_min = thk
        T = analysis.form.thrust()
        swt = analysis.shape.compute_selfweight()
        T_over_swt = T/swt
        n_reduction = n - int(n_step*100)*math.floor(n*100.0/int(n_step*100))/100.0
        print('n (offset) | Min THK  |   Solved  |   Thrust |   T/W   |  n decr. |  Setup time  |   Run time')
        print('{0:.5f}  | {1:.5f}  |   True  |   {2:.1f} |   {3:.6f} |   {4:.4f}   | {5:.2f}s  |   {6:.2f}s'.format(n, thk_min, T, T_over_swt,
                                                                                                                    n_reduction, setup_time, run_time))

        # STORE

        address0 = os.path.join(compas_tno.get('/temp/'), 'form0.json')
        analysis.form.to_json(address0)

        solutions_min.append(T_over_swt)
        solutions_max.append(-1 * T_over_swt)
        thicknesses_min.append(thk)
        thicknesses_max.append(thk)

        # SAVE
        if save_forms:
            address_min = save_forms + '_' + 'min' + '_thk_' + str(100*thk) + '.json'
            address_max = save_forms + '_' + 'max' + '_thk_' + str(100*thk) + '.json'
            analysis.form.to_json(address_min)
            analysis.form.to_json(address_max)

        if n < 0:
            print('Warning: Reduction of cross section (n), or offset is negative:', n_reduction)
            return

        n0 = round(n - n_reduction, 5)
        print('from here we ewill use this n:', n0)
        n_reduction = n_step

        print('Changing solver to IPOPT')
        analysis.optimiser.settings['solver'] == 'IPOPT'

    else:
        print('Error: Minimum THK Optimisation did not find a solution: Try entering a different geometry')
        return

    # Start inverse loop for min/max:

    analysis.optimiser.settings['variables'] = ['ind', 'zb']
    for analysis.optimiser.settings['objective'] in objectives:

        analysis.form = FormDiagram.from_json(address0)

        exitflag = 0
        n = n0
        thk = thk0 - 2 * n
        # print('resulting in this start thickness', thk)
        count = 0
        first_fail = True

        print('\n----- Starting the inverse [', analysis.optimiser.settings['objective'], '] problem for minimum thk:', thk)
        print('n (offset) | THK  |   Solved  |   Opt.Val |   Opt/W   |   THK incr.  |   Setup time  |   Run time')

        while exitflag == 0 and thk <= thk0:

            time0 = time.time()
            intrados = initial_intrados.offset_mesh(n=n, direction='up')
            extrados = initial_extrados.offset_mesh(n=n, direction='down')
            analysis.shape = Shape.from_meshes_and_formdiagram(analysis.form, intrados, extrados, middle=middle, data={'type': 'general', 't': t, 'thk': thk})
            analysis.shape.ro = ro
            swt = analysis.shape.compute_selfweight()

            lumped_swt = analysis.form.lumped_swt()
            analysis.apply_selfweight()
            analysis.form.scale_form(swt/lumped_swt)

            analysis.apply_envelope()
            analysis.apply_reaction_bounds()
            analysis.set_up_optimiser()
            setup_time = time.time() - time0
            analysis.run()
            run_time = time.time() - time0 - setup_time
            exitflag = analysis.optimiser.exitflag  # get info if optimisation was succeded ot not
            fopt = analysis.optimiser.fopt  # objective function optimum value
            fopt_over_weight = fopt/swt  # divide by selfweight

            if exitflag == 0:
                if analysis.optimiser.settings['objective'] == 'min':
                    solutions_min.append(fopt_over_weight)
                    thicknesses_min.append(thk)
                else:
                    solutions_max.append(fopt_over_weight)
                    thicknesses_max.append(thk)
                if save_forms:
                    address = save_forms + '_' + analysis.optimiser.settings['objective'] + '_thk_' + str(100*thk) + '.json'
                else:
                    address = os.path.join(compas_tno.get('/temp/'), 'form_' + analysis.optimiser.settings['objective'] + '.json')
                analysis.form.to_json(address)
                print('{0:.5f}  | {1:.5f}  |   True  |   {2:.1f} |   {3:.6f} |   {4}   | {5:.2f}s  |   {6:.2f}s'.format(
                    n, thk, fopt, fopt_over_weight, n_reduction, setup_time, run_time))
                n = n - n_reduction
                thk = thk0 - 2*n
                first_fail = True
            else:
                print('Failed Optimisation for [{0}] with thk: {1}'.format(analysis.optimiser.settings['objective'], thk))
                other_objective = 'min' if analysis.optimiser.settings['objective'] == 'max' else 'max'
                if thk < thk0:
                    if not first_fail:
                        n = n - n_reduction
                        thk = thk0 - 2*n
                        first_fail = True
                        print('Second Fail with [{0}] opt, will load [{1}] opt in incr. thk: {2}'.format(analysis.optimiser.settings['objective'], other_objective, thk))
                    else:
                        print('First Fail with [{0}] opt, will load [{1}] opt in thk: {2}'.format(analysis.optimiser.settings['objective'], other_objective, thk))
                        first_fail = False
                    address_other_obj = save_forms + '_' + other_objective + '_thk_' + str(100*thk) + '.json'
                    analysis.form = FormDiagram.from_json(address_other_obj)
                    exitflag = 0
            count += 1
        print('---------- End of process -----------', '\n')

    if printout:
        print('------------------ SUMMARY ------------------ ')
        print('ANALYSIS FOUND MIN THK:', thk_min)
        print('thicknesses min ({0}):'.format(len(thicknesses_min)))
        print(thicknesses_min)
        print('thicknesses max ({0}):'.format(len(thicknesses_max)))
        print(thicknesses_max)
        print('min solutions ({0}):'.format(len(solutions_min)))
        print(solutions_min)
        print('max solutions ({0}):'.format(len(solutions_max)))
        print(solutions_max)

    return [thicknesses_min, thicknesses_max],  [solutions_min, solutions_max]



import math

from compas_tno.shapes.dome import dome_ub_lb_update
from compas_tno.shapes.dome import dome_zt_update

from compas_tno.shapes.crossvault import crossvault_ub_lb_update
from compas_tno.shapes.crossvault import crossvault_middle_update

from compas_tno.shapes.pointed_crossvault import pointed_vault_ub_lb_update
from compas_tno.shapes.pointed_crossvault import pointed_vault_middle_update

from compas_tno.shapes.circular_arch import arch_ub_lb_update
from compas_tno.shapes.pointed_arch import pointed_arch_ub_lb_update

from compas_tno.shapes.pavillionvault import pavillionvault_ub_lb_update
from compas_tno.shapes.pavillionvault import pavillionvault_middle_update

from compas_tno.utilities.interpolation import get_shape_ub_pattern
from compas_tno.utilities.interpolation import get_shape_lb_pattern
from compas_tno.utilities.interpolation import get_shape_middle_pattern


__all__ = [
    'apply_envelope_from_shape',
    'apply_envelope_on_xy',
    'apply_envelope_on_xy_from_base',
    'apply_bounds_on_q',
    'project_mesh_to_middle',
    'modify_shapedata_with_spr_angle',
    'apply_envelope_from_shape_proxy',
    'apply_bounds_tub_tlb',
    'apply_bounds_reactions'
]


def apply_envelope_from_shape_proxy(formdata, shapedata):
    """ Apply an envelope (intrados and extrados) to the FormDiagram based on the input shape data

    Parameters
    ----------
    formdata : dict
        The data of the form diagram
    shapedata : dict
        The data of the shape

    Returns
    ----------
    formdata : dict
        The data of the form diagram after assigning the envelope
    """

    from compas_tno.diagrams import FormDiagram
    from compas_tno.shapes import Shape

    form = FormDiagram.from_data(formdata)
    shape = Shape.from_data(shapedata)

    apply_envelope_from_shape(form, shape)

    return form.to_data()


def apply_envelope_from_shape(form, shape):
    """ Apply an envelope (intrados and extrados) to the FormDiagram based on the input shape.

    Parameters
    ----------
    form : FormDiagram
        The input FormDiagram.
    shape : Shape
        The input Shape with intrados and extrardos.

    Returns
    ----------
    None
        The formdiagram is updated in place.
    """

    x = form.vertices_attribute('x')  # check if array is necessary here
    y = form.vertices_attribute('y')

    if shape.datashape['type'] == 'dome':
        zub, zlb = dome_ub_lb_update(x, y, shape.datashape['thk'], shape.datashape['t'], shape.datashape['center'], shape.datashape['radius'])
    elif shape.datashape['type'] == 'crossvault':
        zub, zlb = crossvault_ub_lb_update(x, y, shape.datashape['thk'], shape.datashape['t'], shape.datashape['xy_span'])
    elif shape.datashape['type'] == 'pointed_crossvault':
        zub, zlb = pointed_vault_ub_lb_update(x, y, shape.datashape['thk'], shape.datashape['t'], shape.datashape['xy_span'], hc=shape.datashape['hc'],
                                              he=shape.datashape['he'], hm=shape.datashape['hm'])
    elif shape.datashape['type'] == 'arch':
        zub, zlb = arch_ub_lb_update(x, y, shape.datashape['thk'], shape.datashape['t'], H=shape.datashape['H'], L=shape.datashape['L'], x0=shape.datashape['x0'])
    elif shape.datashape['type'] == 'pointed_arch':
        zub, zlb = pointed_arch_ub_lb_update(x, y, shape.datashape['thk'], shape.datashape['t'], hc=shape.datashape['hc'], L=shape.datashape['L'], x0=shape.datashape['x0'])
    elif shape.datashape['type'] == 'pavillionvault':
        zub, zlb = pavillionvault_ub_lb_update(x, y, shape.datashape['thk'], shape.datashape['t'], shape.datashape['xy_span'], shape.datashape['spr_angle'])
    elif shape.datashape['type'] == 'general':
        XY = form.vertices_attributes('xy')
        zub = get_shape_ub_pattern(shape, XY)
        zlb = get_shape_lb_pattern(shape, XY)
    else:
        raise Exception

    i = 0

    for key in form.vertices():
        ub_ = float(zub[i])
        lb_ = float(zlb[i])
        x, y, _ = form.vertex_coordinates(key)
        form.vertex_attribute(key, 'ub', value=ub_)
        form.vertex_attribute(key, 'lb', value=lb_)
        i += 1

    return


def apply_envelope_on_xy(form, c=0.5):
    """ Apply an envelope to the FormDiagram in the plan (x, y) by a given distance.

    Parameters
    ----------
    form : FormDiagram
        The input FormDiagram.
    c : float, optional
        The maximum allowed movement of the nodes in ``x`` or ``y``, by default 0.5

    Returns
    ----------
    None
        The formdiagram is updated in place in the attributes
    """

    for key, vertex in form.vertex.items():
        form.vertex_attribute(key, 'xmin', vertex.get('x') - c)
        form.vertex_attribute(key, 'xmax', vertex.get('x') + c)
        form.vertex_attribute(key, 'ymin', vertex.get('y') - c)
        form.vertex_attribute(key, 'ymax', vertex.get('y') + c)

    return


def apply_envelope_on_xy_from_base(form, form_base, c=0.5):
    """ Apply an envelope to the FormDiagram considering a given distance applied to a base form diagram.

    Parameters
    ----------
    form : FormDiagram
        The input FormDiagram.
    form_base ::FormDiagram
        The base FormDiagram to consider.
    c : float, optional
        The maximum allowed movement of the nodes in ``x`` or ``y``, by default 0.5

    Returns
    ----------
    None
        The formdiagram is updated in place in the attributes
    """

    for key, vertex in form_base.vertex.items():
        form.vertex_attribute(key, 'xmin', vertex.get('x') - c)
        form.vertex_attribute(key, 'xmax', vertex.get('x') + c)
        form.vertex_attribute(key, 'ymin', vertex.get('y') - c)
        form.vertex_attribute(key, 'ymax', vertex.get('y') + c)

    return


def apply_bounds_on_q(form, qmin=-1e+4, qmax=1e-8):
    """ Apply bounds on the magnitude of the edges'force densities.

    Parameters
    ----------
    form : FormDiagram
        The input FormDiagram
    qmin : float, optional
        The minimum allowed force density ``qmin``, by default -1e+4
    qmax : float, optional
        The maximum allowed force density ``qmax``, by default 1e-8

    Returns
    ----------
    None
        The formdiagram is updated in place in the attributes.
    """

    if isinstance(qmin, list):
        for i, (u, v) in enumerate(form.edges_where({'_is_edge': True})):
            form.edge_attribute((u, v), 'qmin', qmin[i])
            form.edge_attribute((u, v), 'qmax', qmax[i])
    else:
        for u, v in form.edges_where({'_is_edge': True}):
            form.edge_attribute((u, v), 'qmin', qmin)
            form.edge_attribute((u, v), 'qmax', qmax)

    return


def apply_bounds_reactions(form, shape, assume_shape=None):
    """ Apply bounds on the magnitude of the allowed increase in thickness of the upper-bound (tub), lower-bound (tlb), and of the reaction vector (tub_reacmax).

    Parameters
    ----------
    form : FormDiagram
        The input FormDiagram
    shape : Shape
        The shape of masonry to constraint the form diagram to.
    assume_shape : dict, optional
        Whether or not consider the settings of a different shape on the process, by default None

    Returns
    ----------
    None
        The formdiagram is updated in place in the attributes.
    """

    if assume_shape:
        data = assume_shape
    else:
        data = shape.datashape

    thk = data['thk']

    if data['type'] == 'dome' or data['type'] == 'dome_polar':
        [x0, y0] = data['center'][:2]
        for key in form.vertices_where({'is_fixed': True}):
            x, y, _ = form.vertex_coordinates(key)
            theta = math.atan2((y - y0), (x - x0))
            x_ = thk/2*math.cos(theta)
            y_ = thk/2*math.sin(theta)
            form.vertex_attribute(key, 'b', [x_, y_])

    b_manual = data.get('b_manual', None)
    if data['type'] == 'arch':
        H = data['H']
        L = data['L']
        thk = data['thk']
        radius = H / 2 + (L**2 / (8 * H))
        zc = radius - H
        re = radius + thk/2
        x = math.sqrt(re**2 - zc**2)
        for key in form.vertices_where({'is_fixed': True}):
            form.vertex_attribute(key, 'b', [x - L/2, 0.0])
            if b_manual:
                form.vertex_attribute(key, 'b', [b_manual, 0.0])
                print('Applied b manual')

    if data['type'] == 'dome_spr':
        x0 = data['center'][0]
        y0 = data['center'][1]
        radius = data['radius']
        [_, theta_f] = data['theta']
        r_proj_e = (radius + thk/2) * math.sin(theta_f)
        r_proj_m = (radius) * math.sin(theta_f)
        delt = r_proj_e - r_proj_m
        for key in form.vertices_where({'is_fixed': True}):
            x, y, _ = form.vertex_coordinates(key)
            theta = math.atan2((y - y0), (x - x0))
            x_ = delt*math.cos(theta)
            y_ = delt*math.sin(theta)
            form.vertex_attribute(key, 'b', [x_, y_])

    if data['type'] == 'pavillionvault':
        x0, x1 = data['xy_span'][0]
        y0, y1 = data['xy_span'][1]
        for key in form.vertices_where({'is_fixed': True}):
            x, y, _ = form.vertex_coordinates(key)
            if x == x0:
                form.vertex_attribute(key, 'b', [-thk/2, 0])
            elif x == x1:
                form.vertex_attribute(key, 'b', [+thk/2, 0])
            if y == y0:
                if form.vertex_attribute(key, 'b'):
                    b = form.vertex_attribute(key, 'b')
                    b[1] = -thk/2
                    form.vertex_attribute(key, 'b', b)
                else:
                    form.vertex_attribute(key, 'b', [0, -thk/2])
            elif y == y1:
                if form.vertex_attribute(key, 'b'):
                    b = form.vertex_attribute(key, 'b')
                    b[1] = +thk/2
                    form.vertex_attribute(key, 'b', b)
                else:
                    form.vertex_attribute(key, 'b', [0, +thk/2])

    return


def apply_bounds_tub_tlb(form, tubmax=0.5, tlbmax=0.5):
    """ Apply bounds on the magnitude of the allowed increase in thickness of the upper-bound (tub), lower-bound (tlb), and of the reaction vector (tub_reacmax).

    Parameters
    ----------
    form : FormDiagram
        The input FormDiagram
    tubmax : float, optional
        The maximum increase in thickness of the extrados. The default value is ``0.5``.
    tlbmax : float, optional
        The maximum increase in thickness of the intrados. The default value is ``0.5``.

    Returns
    ----------
    None
        The formdiagram is updated in place in the attributes.
    """

    for vertex in form.vertices():
        form.vertex_attribute(vertex, 'tubmax', tubmax)
        form.vertex_attribute(vertex, 'tlbmax', tlbmax)

    return


def project_mesh_to_middle(mesh, shape=None):
    """ Project a mesh to the middle surface of the shape.

    Parameters
    ----------
    mesh : Mesh
        The input Mesh.
    shape : Shape
        The input Shape with a middle surface, by default None.

    Returns
    ----------
    None
        The formdiagram is updated in place in the attributes.
    """

    x = mesh.vertices_attribute('x')
    y = mesh.vertices_attribute('y')

    if shape.datashape['type'] == 'dome':
        zt = dome_zt_update(x, y, shape.datashape['radius'], shape.datashape['t'], shape.datashape['center'])
    elif shape.datashape['type'] == 'crossvault':
        zt = crossvault_middle_update(x, y,  shape.datashape['t'],  xy_span=shape.datashape['xy_span'])
    elif shape.datashape['type'] == 'pointed_crossvault':
        zt = pointed_vault_middle_update(x, y,  shape.datashape['t'],  xy_span=shape.datashape['xy_span'],
                                         hc=shape.datashape['hc'], he=shape.datashape['he'], hm=shape.datashape['hm'])
    elif shape.datashape['type'] == 'pavillionvault':
        zt = pavillionvault_middle_update(x, y,  shape.datashape['t'],  xy_span=shape.datashape['xy_span'], spr_angle=shape.datashape['spr_angle'])
    else:
        XY = mesh.vertices_attributes('xy')
        zt = get_shape_middle_pattern(shape, XY)

    i = 0

    for key in mesh.vertices():
        z = float(zt[i])
        x, y, _ = mesh.vertex_coordinates(key)
        mesh.vertex_attribute(key, 'z', value=z)
        i += 1

    return


def modify_shapedata_with_spr_angle(datashape):
    """ Modify the Shape data to account for an springing angle.
    This is done by increasing the ``xy_span`` limits and works exclusivvely for rectandular diagrams.

    Parameters
    ----------
    datashape : dict
        The data to construct the Shape.

    Returns
    ----------
    datashape : dict
        The updated datashape.
    """

    typevault = datashape.get('type')
    spr_angle = datashape.get('spr_angle', 0.0)

    if spr_angle:
        if typevault in ['crossvault', 'pointed_crossvault', 'pavillionvault', ':parabolic_shell', 'domicalvault']:
            xy_span = datashape['xy_span']
            span_x = xy_span[0][1] - xy_span[0][0]
            span_y = xy_span[1][1] - xy_span[1][0]

            A = 1/math.cos(math.radians(spr_angle))
            print('deg, discretisation:', A, spr_angle)
            xy_span_enlarged = [[-span_x/2*(A - 1), span_x*(1 + (A - 1)/2)], [-span_y/2*(A - 1), span_y*(1 + (A - 1)/2)]]

            datashape['xy_span'] = xy_span_enlarged
        else:
            raise NotImplementedError

    return datashape



"""
********************************************************************************
compas_tno.utilities
********************************************************************************

.. currentmodule:: compas_tno.utilities

Symmetry
====================

.. autosummary::
    :toctree: generated/

    apply_radial_symmetry
    apply_symmetry_from_axis
    find_sym_axis_in_rect_patterns
    build_symmetry_matrix
    build_symmetry_transformation
    build_vertex_symmetry_transformation
    build_symmetry_matrix_supports

Constraints
====================

.. autosummary::
    :toctree: generated/

    check_envelope_constraints
    distance_target
    rectangular_smoothing_constraints
    assign_cracks
    rollers_on_openings
    set_b_constraint
    set_rollers_constraint

Loads
====================

.. autosummary::
    :toctree: generated/

    apply_selfweight_from_shape
    apply_selfweight_from_pattern
    apply_horizontal_multiplier
    apply_fill_load
    apply_selfweight_from_shape_proxy
    apply_selfweight_from_thrust


Envelopes
=========

.. autosummary::
    :toctree: generated/

    apply_envelope_from_shape
    apply_envelope_on_xy
    apply_envelope_on_xy_from_base
    apply_bounds_on_q
    project_mesh_to_middle
    modify_shapedata_with_spr_angle
    apply_envelope_from_shape_proxy
    apply_bounds_tub_tlb
    apply_bounds_reactions

Interpolation
====================

.. autosummary::
    :toctree: generated/

    interpolate_from_pointcloud
    get_shape_ub
    get_shape_ub_pattern
    get_shape_ub_fill
    get_shape_lb
    get_shape_lb_pattern
    get_shape_middle
    delaunay_mesh_from_points
    mesh_from_pointcloud
    create_mesh_from_topology_and_pointcloud
    create_mesh_from_topology_and_basemesh

Stiffness
====================

.. autosummary::
    :toctree: generated/

    compute_form_initial_lengths
    compute_edge_stiffness
    compute_average_edge_stiffness

Blocks
======

.. autosummary::
    :toctree: generated/

    blocks_from_dual
    extended_dual

Form Modifications
====================

.. autosummary::
    :toctree: generated/

    split_intersection_lines
    form_add_lines_support
    form_parabolic_slide
    move_pattern_to_origin
    slide_diagram
    mesh_remove_two_valent_nodes
    slide_pattern_inwards
    displacement_map_paraboloid
    displacement_map_4parabolas
    displacement_map_parabola
    shuffle_diagram

Data Analysis
====================

.. autosummary::
    :toctree: generated/

    diagram_of_thrust
    diagram_of_multiple_thrust
    diagram_of_thrust_load_mult
    surface_GSF_load_mult
    save_csv_row
    open_csv_row
    interpolate_min_thk
    filter_min_thk
    lookup_folder
    save_pointcloud


"""
from __future__ import absolute_import

from .symmetry import (
    apply_radial_symmetry,
    apply_symmetry_from_axis,
    find_sym_axis_in_rect_patterns,
    build_symmetry_matrix,
    build_symmetry_transformation,
    build_vertex_symmetry_transformation,
    build_symmetry_matrix_supports,
)

from .constraints import (
    check_envelope_constraints,
    distance_target,
    rectangular_smoothing_constraints,
    assign_cracks,
    rollers_on_openings,
    set_b_constraint,
    set_rollers_constraint
)

from .functions import (
    paraboloid,
    dome
)

from .loads import (
    apply_selfweight_from_shape,
    apply_selfweight_from_pattern,
    apply_selfweight_from_thrust,
    apply_horizontal_multiplier,
    apply_fill_load,
    apply_selfweight_from_shape_proxy

)

from .envelopes import (
    apply_envelope_from_shape,
    apply_envelope_on_xy,
    apply_envelope_on_xy_from_base,
    apply_bounds_on_q,
    project_mesh_to_middle,
    modify_shapedata_with_spr_angle,
    apply_envelope_from_shape_proxy,
    apply_bounds_tub_tlb,
    apply_bounds_reactions
)

from .interpolation import (
    interpolate_from_pointcloud,
    get_shape_ub,
    get_shape_ub_pattern,
    get_shape_ub_fill,
    get_shape_lb,
    get_shape_lb_pattern,
    get_shape_middle,
    delaunay_mesh_from_points,
    mesh_from_pointcloud,
    create_mesh_from_topology_and_pointcloud,
    create_mesh_from_topology_and_basemesh
)

from .stiffness import (
    compute_form_initial_lengths,
    compute_edge_stiffness,
    compute_average_edge_stiffness
)

from .json import (
    update_json
)

from .blocks import (
    blocks_from_dual,
    extended_dual
)

from .form import (
    split_intersection_lines,
    form_add_lines_support,
    form_parabolic_slide,
    move_pattern_to_origin,
    slide_diagram,
    mesh_remove_two_valent_nodes,
    store_inds,
    slide_pattern_inwards,
    displacement_map_paraboloid,
    displacement_map_4parabolas,
    displacement_map_parabola,
    shuffle_diagram
)

from .data_analysis import (
    diagram_of_thrust,
    diagram_of_multiple_thrust,
    diagram_of_thrust_load_mult,
    surface_GSF_load_mult,
    save_csv_row,
    open_csv_row,
    interpolate_min_thk,
    filter_min_thk,
    lookup_folder,
    save_pointcloud
)


__all__ = [

    'apply_radial_symmetry',
    'apply_symmetry_from_axis',
    'find_sym_axis_in_rect_patterns',
    'build_symmetry_matrix',
    'build_symmetry_transformation',
    'build_vertex_symmetry_transformation',
    'build_symmetry_matrix_supports',

    'check_envelope_constraints',
    'distance_target',
    'rectangular_smoothing_constraints',
    'assign_cracks',
    'rollers_on_openings',
    'set_b_constraint',
    'set_rollers_constraint',

    'paraboloid',
    'dome',

    'apply_selfweight_from_shape',
    'apply_selfweight_from_pattern',
    'apply_selfweight_from_thrust',
    'apply_horizontal_multiplier',
    'apply_fill_load',

    'apply_selfweight_from_shape_proxy',

    'apply_envelope_from_shape',
    'apply_envelope_on_xy',
    'apply_envelope_on_xy_from_base',
    'apply_bounds_on_q',
    'project_mesh_to_middle',
    'modify_shapedata_with_spr_angle',
    'apply_envelope_from_shape_proxy',
    'apply_bounds_tub_tlb',
    'apply_bounds_reactions',

    'interpolate_from_pointcloud',
    'get_shape_ub',
    'get_shape_ub_pattern',
    'get_shape_ub_fill',
    'get_shape_lb',
    'get_shape_lb_pattern',
    'get_shape_middle',
    'delaunay_mesh_from_points',
    'mesh_from_pointcloud',
    'create_mesh_from_topology_and_pointcloud',
    'create_mesh_from_topology_and_basemesh',

    'compute_form_initial_lengths',
    'compute_edge_stiffness',
    'compute_average_edge_stiffness',

    'update_json',

    'blocks_from_dual',
    'extended_dual',

    'split_intersection_lines',
    'form_add_lines_support',
    'form_parabolic_slide',
    'move_pattern_to_origin',
    'slide_diagram',
    'mesh_remove_two_valent_nodes',
    'store_inds',
    'slide_pattern_inwards',
    'displacement_map_paraboloid',
    'displacement_map_4parabolas',
    'displacement_map_parabola',
    'shuffle_diagram',

    'diagram_of_thrust',
    'diagram_of_multiple_thrust',
    'diagram_of_thrust_load_mult',
    'surface_GSF_load_mult',
    'save_csv_row',
    'open_csv_row',
    'interpolate_min_thk',
    'filter_min_thk',
    'lookup_folder',
    'save_pointcloud',
]



import cyipopt

import time

from numpy import hstack
from numpy import array

try:
    from torch import tensor

    from compas_tno.autodiff.equilibrium_pytorch import f_constraints_pytorch
    from compas_tno.autodiff.equilibrium_pytorch import f_objective_pytorch
    from compas_tno.autodiff.equilibrium_pytorch import compute_autograd
    from compas_tno.autodiff.equilibrium_pytorch import compute_autograd_jacobian
except BaseException:
    pass  # Module tensor not available

from .post_process import post_process_general


class Wrapper_ipopt_autodiff(object):
    """Wrapper to send to IPOPT using autodifferentiation
    """
    def __init__(self):
        self.fobj = None
        self.fconstr = None
        self.args_obj = None
        self.args_constr = None
        self.bounds = None
        self.x0 = None
        self.eps = 1e-8
        pass

    def objective(self, x):
        """The callback for calculating the objective

        Parameters
        ----------
        x : array
            The variabless passed to IPOPT

        Returns
        -------
        fopt : float
            The objective function value at x.
        """

        variables = tensor(x.reshape(-1, 1))
        return array(self.fobj(variables, *self.args_obj))

    def gradient(self, x):
        """The callback for calculating the gradient

        Parameters
        ----------
        x : array
            The variabless passed to IPOPT

        Returns
        -------
        grad : array
            The gradient of the objective function at x.
        """

        variables = tensor(x.reshape(-1, 1), requires_grad=True)
        f = self.fobj(variables, *self.args_obj)
        return array(compute_autograd(variables, f))

    def constraints(self, x):
        """The callback for calculating the constraints

        Parameters
        ----------
        x : array
            The variabless passed to IPOPT

        Returns
        -------
        constr : array
            The constraints of the objective function at x.
        """

        variables = tensor(x.reshape(-1, 1))
        return array(self.fconstr(variables, *self.args_constr))

    def jacobian(self, x):
        """The callback for calculating the jacobian

        Parameters
        ----------
        x : array
            The variabless passed to IPOPT

        Returns
        -------
        jac : array
            The gradient of the jacobian matrix at x.
        """
        variables = tensor(x.reshape(-1, 1), requires_grad=True)
        constraints = self.fconstr(variables, *self.args_constr)
        return array(compute_autograd_jacobian(variables, constraints)).flatten()


class Wrapper_ipopt(object):
    """Wrapper to send to IPOPT using analytical derivatives
    """
    def __init__(self):
        self.fobj = None
        self.fconstr = None
        self.args = None
        self.fgrad = None
        self.bounds = None
        self.callback = None
        self.x0 = None
        self.eps = 1e-8
        self.fjac = None
        pass

    def objective(self, x):
        """The callback for calculating the objective

        Parameters
        ----------
        x : array
            The variabless passed to IPOPT

        Returns
        -------
        fopt : float
            The objective function value at x.
        """

        if self.callback:
            self.callback(x, *self.args)
        return self.fobj(x, *self.args)

    def gradient(self, x):
        """The callback for calculating the gradient

        Parameters
        ----------
        x : array
            The variabless passed to IPOPT

        Returns
        -------
        grad : array
            The gradient of the objective function at x.
        """

        return self.fgrad(x, *self.args)

    def constraints(self, x):
        """The callback for calculating the constraints

        Parameters
        ----------
        x : array
            The variabless passed to IPOPT

        Returns
        -------
        constr : array
            The constraints of the objective function at x.
        """

        return self.fconstr(x, *self.args).reshape(-1, 1)

    def jacobian(self, x):
        """The callback for calculating the jacobian

        Parameters
        ----------
        x : array
            The variabless passed to IPOPT

        Returns
        -------
        jac : array
            The gradient of the jacobian matrix at x.
        """

        return self.fjac(x, *self.args).flatten()


def run_optimisation_ipopt(analysis):
    """ Run nonlinear optimisation problem with IPOPT

    Parameters
    ----------
    obj : analysis
        Analysis object with information about optimiser, form and shape.

    Returns
    -------
    obj : analysis
        Analysis object optimised.

    """

    optimiser = analysis.optimiser

    constraints = optimiser.settings['constraints']
    objective = optimiser.settings['objective']
    printout = optimiser.settings.get('printout', False)
    gradients = optimiser.settings.get('gradient', False)
    variables = optimiser.settings['variables']
    callback = optimiser.callback

    bounds = optimiser.bounds
    x0 = optimiser.x0
    g0 = optimiser.g0
    args = [optimiser.M]

    lower = [lw[0] for lw in bounds]
    upper = [up[1] for up in bounds]

    # Tensor modification

    if not gradients:

        (q, ind, dep, E, Edinv, Ei, C, Ct, Ci, Cit, Cf, U, V, p, px, py, pz, z, free, fixed, lh, sym, k, lb, ub, lb_ind, ub_ind, s, Wfree, x, y, b, joints, cracks_lb, cracks_ub,
         free_x, free_y, rol_x, rol_y, Citx, City, Cftx, Cfty, qmin, constraints, max_rol_rx, max_rol_ry, Asym) = args[:48]

        EdinvEi = Edinv*Ei
        Edinv_p = Edinv.dot(p)

        EdinvEi_th = tensor(EdinvEi)
        Edinv_p_th = tensor(Edinv_p)
        C_th = tensor(C.toarray())
        Ci_th = tensor(Ci.toarray())
        Cit_th = Ci_th.t()
        Cf_th = tensor(Cf.toarray())
        pzfree = tensor(pz[free])
        xyz = tensor(hstack([x, y, z]))
        xy = tensor(hstack([x, y]))
        pfixed = tensor(hstack([px, py, pz])[fixed])
        U_th = tensor(U.toarray())
        V_th = tensor(V.toarray())

        args_obj = (Edinv_p_th, EdinvEi_th, ind, dep, C_th, Ci_th, Cit_th, Cf_th, pzfree, xyz, xy, pfixed, k, objective)
        args_constr = (Edinv_p_th, EdinvEi_th, ind, dep, C_th, Ci_th, Cit_th, Cf_th, pzfree, xyz, xy, pfixed, k, free, fixed,
                       ub, lb, ub_ind, lb_ind, b, constraints, max_rol_rx, max_rol_ry, rol_x, rol_y, px, py, Asym, U_th, V_th)

        problem_obj = Wrapper_ipopt_autodiff()
        problem_obj.fobj = f_objective_pytorch
        problem_obj.fconstr = f_constraints_pytorch
        problem_obj.args_obj = args_obj
        problem_obj.args_constr = args_constr
        problem_obj.bounds = bounds
        problem_obj.x0 = x0
        problem_obj.args = args

        variables = tensor(x0, requires_grad=True)
        g0 = f_constraints_pytorch(variables, *args_constr)

    else:

        problem_obj = Wrapper_ipopt()
        problem_obj.fobj = optimiser.fobj
        problem_obj.fconstr = optimiser.fconstr
        problem_obj.fjac = optimiser.fjac
        problem_obj.args = args
        problem_obj.fgrad = optimiser.fgrad
        problem_obj.bounds = bounds
        problem_obj.callback = callback
        problem_obj.x0 = x0

        if printout:
            g0 = optimiser.fconstr(x0, *args)

    cu = [10e10]*len(g0)
    cl = [0.0]*len(g0)
    # if any(el in ['symmetry', 'symmetry-horizontal', 'symmetry-vertical'] for el in constraints):
    #     nsym = Asym.shape[0]
    #     cu[-nsym:] = [0.0]*nsym
    #     cl[-nsym:] = [0.0]*nsym

    nlp = cyipopt.Problem(
        n=len(x0),
        m=len(g0),
        problem_obj=problem_obj,
        lb=lower,
        ub=upper,
        cl=cl,
        cu=cu
    )

    # Set Options and Time
    nlp = _nlp_options(nlp, optimiser)
    start_time = time.time()

    # Solve
    xopt, info = nlp.solve(x0)
    fopt = info['obj_val']
    exitflag = info['status']
    if exitflag == 1 or exitflag == 0:  # IPOPT consider solved = 1. Solved in tolerances 0 -> TNO solved = 0
        exitflag = 0
    else:
        exitflag = 1
    if printout:
        print(info['status_msg'])

    elapsed_time = time.time() - start_time
    if printout:
        print('Solving Time: {0:.1f} sec'.format(elapsed_time))

    optimiser.exitflag = exitflag
    optimiser.time = elapsed_time
    optimiser.fopt = fopt
    optimiser.xopt = xopt
    optimiser.niter = None  # Did not find a way to display number of iterations
    optimiser.message = info['status_msg']
    optimiser.nlp = nlp

    post_process_general(analysis)

    return analysis


def _nlp_options(nlp, optimiser):
    """Set NLP options for IPOPT

    Parameters
    ----------
    nlp : obj
        IPOPT object
    optimiser : Optimiser
        The Optimiser object

    Returns
    -------
    [type]
        [description]
    """

    # Link to instructions: https://coin-or.github.io/Ipopt/OPTIONS.html

    # nlp.add_option(b'hessian_approximation', b'limited-memory')
    # nlp.add_option('tol', 1e-4)              # Default 1e-8
    # nlp.add_option('max_iter', 500)        # Default 3000

    # nlp.add_option('dual_inf_tol', 100.0)  # Default 1.0
    # nlp.add_option('constr_viol_tol', 0.1) # Default 1e-4
    # nlp.add_option('compl_inf_tol', 0.1)   # Default 1e-4

    # nlp.add_option('acceptable_iter', 10)
    # nlp.add_option('acceptable_tol', 1e-3)
    # nlp.add_option('acceptable_constr_viol_tol', 1e-4)  # Default 1e-2
    # nlp.add_option('acceptable_dual_inf_tol', 10e10)  # Default 10e10
    # nlp.add_option('acceptable_compl_inf_tol', 1e-2)  # Default 1e-2
    # nlp.add_option('max_iter', 500)

    scaling = optimiser.settings.get('nlp_scaling_method', None)

    if not optimiser.settings['printout']:
        nlp.add_option('print_level', 0)
    if optimiser.settings.get('derivative_test', None):
        nlp.add_option('derivative_test', 'first-order')
        # nlp.add_option('derivative_test_perturbation') #, 10e-8
        # nlp.add_option('derivative_test_print_all', 'yes')
    if optimiser.settings.get('max_iter', None):
        nlp.add_option('max_iter', optimiser.settings['max_iter'])
    if scaling:
        print('Applied sollver scaling:', scaling)
        nlp.add_option('nlp_scaling_method', scaling)

    return nlp



from numpy import zeros
from numpy import array
from numpy import ones
import math

from compas_tno.shapes import MeshDos
from compas.datastructures import mesh_delete_duplicate_vertices


def dome_heightfields_proxy(center=[5.0, 5.0], radius=5.0, thk=0.30, t=5.0, discretisation=[8, 20], *args, **kwargs):
    """Function that computes the highfield of a dome through the proxy

    Parameters
    ----------
    center : [float, float], optional
        xy-span of the pattern, by default [[0.0, 10.0], [0.0, 10.0]]
    thk : float, optional
        The thickness of the vault, by default 0.5
    tol : float, optional
        Tolerance, by default 10e-6
    t : float, optional
        Parameter for lower bound in nodes in the boundary, by default 0.0
    discretisation : list|int, optional
        LEvel of discretisation of the shape, by default [100, 100]

    Returns
    -------
    intradosdata
        Data to a Mesh for the intrados of the pattern
    extradosdata
        Data to a Mesh for the extrados of the pattern
    middledata
        Data to a Mesh for the middle of the pattern
    """
    intrados, extrados, middle = set_dome_heighfield(center=center, radius=radius, thk=thk, t=t, discretisation=discretisation)
    return intrados.to_data(), extrados.to_data(), middle.to_data()


def set_dome_heighfield(center=[5.0, 5.0], radius=5.0, thk=0.30, t=0.0, discretisation=[8, 20], expanded=False):
    """Height of the dome heighfield

    Parameters
    ----------
    center : [float, float], optional
        x, y coordinates of the center of the dome, by default [5.0, 5.0]
    radius : float, optional
        The radius of the dome, by default 5.0
    thk : float, optional
        The thickness of the dome, by default 0.30
    t : float, optional
        Parameter to the lowerbound of the nodes in the boundary, by default 0.0
    discretisation : [float, float], optional
        Discretisation level for the dome (hoops and meridians), by default [8, 20]
    expanded : bool, optional
        If should expand 1 row to "close" the dome, by default False

    Returns
    -------
    intrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the intrados of the shape
    extrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the extrados of the shape
    middle : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the middle of the shape
    """

    tol = 10e-3
    xc = center[0]
    yc = center[1]
    ri = radius - thk/2
    re = radius + thk/2
    n_radial = discretisation[0]
    n_spikes = discretisation[1]
    r_div = radius/n_radial
    theta = 2*math.pi/n_spikes

    x1d = []
    y1d = []
    zt1d = []
    zi1d = []
    ze1d = []
    faces = []
    faces_i = []
    i = 0
    uv_i = {}

    for nr in range(n_radial+1):
        for nc in range(n_spikes):
            uv_i[(nr, nc)] = i
            xi = xc + nr * r_div * math.cos(theta * nc)
            yi = yc + nr * r_div * math.sin(theta * nc)
            zt2 = radius**2 - (xi - xc)**2 - (yi - yc)**2
            zi2 = ri**2 - (xi - xc)**2 - (yi - yc)**2
            ze2 = re**2 - (xi - xc)**2 - (yi - yc)**2
            ze = math.sqrt(ze2)
            if zt2 < 0.0:
                zt = 0.0
            else:
                zt = math.sqrt(zt2)
            if zi2 < 0.0:
                if zi2 < -tol:
                    zi = -t
                else:
                    zi = 0.0
            else:
                zi = math.sqrt(zi2)

            x1d.append(xi)
            y1d.append(yi)
            zt1d.append(zt)
            zi1d.append(zi)
            ze1d.append(ze)

            if nr < n_radial:
                p1 = (nr, nc)
                p2 = (nr, nc+1)
                p3 = (nr+1, nc)
                p4 = (nr+1, nc+1)
                if nc < n_spikes - 1:  # General Case
                    # face = [p1, p2, p4, p3]
                    face = [p3, p4, p2, p1]
                else:
                    p2 = (nr, 0)
                    p4 = (nr+1, 0)
                    # face = [p1, p2, p4, p3]
                    face = [p3, p4, p2, p1]
                faces.append(face)

            i += 1

    if expanded:
        nr = n_radial + 1
        faces_exp = faces[:]
        x1d_exp = x1d[:]
        y1d_exp = y1d[:]
        for nc in range(n_spikes):
            uv_i[(nr, nc)] = i
            ze = -t
            zi = -t
            xi = xc + re * math.cos(theta * nc)
            yi = yc + re * math.sin(theta * nc)
            x1d_exp.append(xi)
            y1d_exp.append(yi)
            zi1d.append(zi)
            ze1d.append(ze)

            p1 = (nr-1, nc)
            p2 = (nr-1, nc+1)
            p3 = (nr, nc)
            p4 = (nr, nc+1)
            if nc < n_spikes - 1:  # General Case
                face = [p3, p4, p2, p1]
            else:
                p2 = (nr-1, 0)
                p4 = (nr, 0)
                face = [p3, p4, p2, p1]
            faces_exp.append(face)

            i += 1

    for face in faces:
        face_i = []
        for uv in face:
            u, v = uv
            i = uv_i[(u, v)]
            face_i.append(i)
        faces_i.append(face_i)

    xyz_middle = array([x1d, y1d, zt1d]).transpose()
    middle = MeshDos.from_vertices_and_faces(xyz_middle, faces_i)

    if expanded:
        faces_i_exp = []
        for face in faces_exp:
            face_i = []
            for uv in face:
                u, v = uv
                i = uv_i[(u, v)]
                face_i.append(i)
            faces_i_exp.append(face_i)

        xyz_middle = array([x1d, y1d, zt1d]).transpose()
        xyz_intrados = array([x1d_exp, y1d_exp, zi1d]).transpose()
        xyz_extrados = array([x1d_exp, y1d_exp, ze1d]).transpose()
        middle = MeshDos.from_vertices_and_faces(xyz_middle, faces_i)
        intrados = MeshDos.from_vertices_and_faces(xyz_intrados, faces_i_exp)
        extrados = MeshDos.from_vertices_and_faces(xyz_extrados, faces_i_exp)
    else:
        xyz_middle = array([x1d, y1d, zt1d]).transpose()
        xyz_intrados = array([x1d, y1d, zi1d]).transpose()
        xyz_extrados = array([x1d, y1d, ze1d]).transpose()
        middle = MeshDos.from_vertices_and_faces(xyz_middle, faces_i)
        intrados = MeshDos.from_vertices_and_faces(xyz_intrados, faces_i)
        extrados = MeshDos.from_vertices_and_faces(xyz_extrados, faces_i)

    mesh_delete_duplicate_vertices(middle)  # this is a hack. Do it better
    mesh_delete_duplicate_vertices(intrados)
    mesh_delete_duplicate_vertices(extrados)

    return intrados, extrados, middle


def set_dome_with_spr(center=[5.0, 5.0], radius=5.0, thk=0.30, theta=[0, math.pi/2], t=5.0, discretisation=[8, 20]):
    """Height of the dome heighfield with spring angle

    Parameters
    ----------
    center : [float, float], optional
        x, y coordinates of the center of the dome, by default [5.0, 5.0]
    radius : float, optional
        The radius of the dome, by default 5.0
    thk : float, optional
        The thickness of the dome, by default 0.30
    theta : [float, float], optional
        The range to cover, by default [0, pi/2] which is a hemispheric dome
    t : float, optional
        Parameter to the lowerbound of the nodes in the boundary, by default 0.0
    discretisation : [float, float], optional
        Discretisation level for the dome (hoops and meridians), by default [8, 20]

    Returns
    -------
    intrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the intrados of the shape
    extrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the extrados of the shape
    middle : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the middle of the shape
    """

    center.append(0.0)
    ri = radius - thk/2
    re = radius + thk/2
    phi_lower = 0
    phi_upper = 2 * math.pi
    phi_length = discretisation[1]
    phi_range = [phi_lower + x * (phi_upper - phi_lower) / phi_length for x in range(phi_length + 1)]

    theta_length = discretisation[0]
    theta_lower = theta[0]
    theta_upper = theta[1]
    theta_range = [theta_lower + x * (theta_upper - theta_lower) / theta_length for x in range(theta_length + 1)]

    [xmin, _, _] = geom_dome(center, radius, theta_range[0], phi_range[0])
    [xmax, _, _] = geom_dome(center, radius, theta_range[theta_length], phi_range[phi_length])
    theta_upper_i = (math.pi/2 - math.acos(min(1, (xmax-center[0])/ri)))
    theta_lower_e = (math.pi/2 - math.acos((xmin-center[0])/re))
    # print('min/max', xmin, xmax)
    # print('lower/lower_i/upper/upper_e', theta_lower/(math.pi/2), theta_lower_e/(math.pi/2), theta_upper/(math.pi/2), theta_upper_i/(math.pi/2))
    theta_range_i = [theta_lower + x * (theta_upper_i - theta_lower) / theta_length for x in range(theta_length + 1)]
    theta_range_e = [theta_lower_e + x * (theta_upper - theta_lower_e) / theta_length for x in range(theta_length + 1)]

    faces = []
    faces_i = []
    count = 0
    uv_i = {}
    xyz_middle = []
    xyz_intrados = []
    xyz_extrados = []

    for i in range(phi_length + 1):
        for j in range(theta_length + 1):
            uv_i[(i, j)] = count
            p1 = (i, j)
            p2 = (i, j+1)
            p3 = (i+1, j)
            p4 = (i+1, j+1)
            face = []

            ptm = geom_dome(center, radius, theta_range[j], phi_range[i])
            pti = geom_dome(center, ri, theta_range_i[j], phi_range[i])
            pte = geom_dome(center, re, theta_range_e[j], phi_range[i])

            xyz_middle.append(ptm)
            xyz_intrados.append(pti)
            xyz_extrados.append(pte)

            if i < phi_length and j < theta_length:
                face = [p3, p4, p2, p1]
                faces.append(face)

            count += 1

    for face in faces:
        face_i = []
        for uv in face:
            u, v = uv
            i = uv_i[(u, v)]
            face_i.append(i)
        faces_i.append(face_i)

    xyz_middle = array(xyz_middle)
    xyz_extrados = array(xyz_extrados)
    xyz_intrados = array(xyz_intrados)

    middle = MeshDos.from_vertices_and_faces(xyz_middle, faces_i)
    extrados = MeshDos.from_vertices_and_faces(xyz_extrados, faces_i)
    intrados = MeshDos.from_vertices_and_faces(xyz_intrados, faces_i)

    mesh_delete_duplicate_vertices(middle)  # this is a hack. Do it better
    mesh_delete_duplicate_vertices(intrados)
    mesh_delete_duplicate_vertices(extrados)

    return intrados, extrados, middle


def set_dome_polar_coord(center=[5.0, 5.0], radius=5.0, thk=0.30, theta=[0, math.pi/2], t=0.0, discretisation=[8, 20]):
    """Height of the dome heighfield assuming polar coordinates

    Parameters
    ----------
    center : [float, float], optional
        x, y coordinates of the center of the dome, by default [5.0, 5.0]
    radius : float, optional
        The radius of the dome, by default 5.0
    thk : float, optional
        The thickness of the dome, by default 0.30
    theta : [float, float], optional
        The range to cover, by default [0, pi/2] which is a hemispheric dome
    t : float, optional
        Parameter to the lowerbound of the nodes in the boundary, by default 0.0
    discretisation : [float, float], optional
        Discretisation level for the dome (hoops and meridians), by default [8, 20]

    Returns
    -------
    intrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the intrados of the shape
    extrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the extrados of the shape
    middle : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the middle of the shape
    """

    center0 = center
    center.append(0.0)
    ri = radius - thk/2
    re = radius + thk/2
    phi_lower = 0
    phi_upper = 2 * math.pi
    phi_length = discretisation[1]
    # step_phi = (phi_upper - phi_lower)
    phi_range = [phi_lower + x * (phi_upper - phi_lower) / phi_length for x in range(phi_length + 1)]

    theta_length = discretisation[0]
    theta_lower = theta[0]
    theta_upper = theta[1]
    theta_range = [theta_lower + x * (theta_upper - theta_lower) / theta_length for x in range(theta_length + 1)]

    [xmin, _, _] = geom_dome(center, radius, theta_range[0], phi_range[0])
    [xmax, _, _] = geom_dome(center, radius, theta_range[theta_length], phi_range[phi_length])
    theta_upper_i = (math.pi/2 - math.acos(min(1, (xmax-center[0])/ri)))
    theta_lower_e = (math.pi/2 - math.acos((xmin-center[0])/re))
    # print('min/max', xmin, xmax)
    # print('lower/lower_i/upper/upper_e', theta_lower/(math.pi/2), theta_lower_e/(math.pi/2), theta_upper/(math.pi/2), theta_upper_i/(math.pi/2))
    theta_range_i = [theta_lower + x * (theta_upper_i - theta_lower) / theta_length for x in range(theta_length + 1)]
    theta_range_e = [theta_lower_e + x * (theta_upper - theta_lower_e) / theta_length for x in range(theta_length + 1)]

    faces = []
    faces_i = []
    count = 0
    uv_i = {}
    xyz_middle = []
    xyz_intrados = []
    xyz_extrados = []

    for i in range(phi_length + 1):
        for j in range(theta_length + 2):
            uv_i[(i, j)] = count
            p1 = (i, j)
            p2 = (i, j+1)
            p3 = (i+1, j)
            p4 = (i+1, j+1)
            face = []

            if j <= theta_length:
                ptm = geom_dome(center, radius, theta_range[j], phi_range[i])
                pte = geom_dome(center, re, theta_range_e[j], phi_range[i])
                pti = geom_dome(center, ri, theta_range_i[j], phi_range[i])
            else:
                ptm = geom_dome(center, radius, theta_range[-1], phi_range[i])
                pte = geom_dome(center, re, theta_range_e[-1], phi_range[i])
                pti[:] = ptm
                pti[2] = -t

            xyz_middle.append(ptm)
            xyz_intrados.append(pti)
            xyz_extrados.append(pte)

            if i < phi_length and j < theta_length + 1:
                face = [p3, p4, p2, p1]
                faces.append(face)

            count += 1

    for face in faces:
        face_i = []
        for uv in face:
            u, v = uv
            i = uv_i[(u, v)]
            face_i.append(i)
        faces_i.append(face_i)

    xyz_middle = array(xyz_middle)
    xyz_extrados = array(xyz_extrados)
    xyz_intrados = array(xyz_intrados)

    middle = MeshDos.from_vertices_and_faces(xyz_middle, faces_i)
    extrados = MeshDos.from_vertices_and_faces(xyz_extrados, faces_i)
    intrados = MeshDos.from_vertices_and_faces(xyz_intrados, faces_i)

    for mesh in [middle, intrados, extrados]:
        mesh_delete_duplicate_vertices(mesh)  # this is a hack. Do it better

    for mesh in [middle, intrados, extrados]:
        delete_fkey = []
        for fkey in mesh.faces():
            if len(mesh.face_coordinates(fkey)) == 2:
                delete_fkey.append(fkey)
        for fkey in delete_fkey:
            mesh.delete_face(fkey)

    center = center0

    return intrados, extrados, middle


def geom_dome(p0, ro, theta, phi):
    """Geometry of the dome in polar coordinates

    Parameters
    ----------
    p0 : list
        center of the dome
    ro : float
        ro parameter
    theta : float
        theta parameter
    phi : float
        phi parameter

    Returns
    -------
    point : list
        List of cartesian coordinates
    """
    x = ro * math.sin(theta) * math.cos(phi)
    y = ro * math.sin(theta) * math.sin(phi)
    z = ro * math.cos(theta)
    point = [p0[0] + x, p0[1] + y, p0[2] + z]
    return point


def dome_zt_update(x, y, radius, t, center=[5.0, 5.0]):
    """Update middle of the dome based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    radius : float, optional
        The radius of the dome, by default 5.0
    t : float
        Parameter for lower bound in nodes in the boundary
    center : [float, float], optional
        x, y coordinates of the center of the dome, by default [5.0, 5.0]

    Returns
    -------
    ub : array
        Values of the upper bound in the points
    lb : array
        Values of the lower bound in the points
    """

    xc = center[0]
    yc = center[1]
    zt = ones((len(x), 1))

    for i in range(len(x)):
        zt2 = radius**2 - (x[i] - xc)**2 - (y[i] - yc)**2
        if zt2 > 0:
            zt[i] = math.sqrt(zt2)
        else:
            zt[i] = 0.0

    return zt


def dome_ub_lb_update(x, y, thk, t, center=[5.0, 5.0], radius=5.0):
    """Update upper and lower bounds of the dome based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the dome
    t : float
        Parameter for lower bound in nodes in the boundary
    center : [float, float], optional
        x, y coordinates of the center of the dome, by default [5.0, 5.0]
    radius : float, optional
        The radius of the dome, by default 5.0

    Returns
    -------
    ub : array
        Values of the upper bound in the points
    lb : array
        Values of the lower bound in the points
    """

    xc = center[0]
    yc = center[1]
    ri = radius - thk/2
    re = radius + thk/2
    ub = ones((len(x), 1))
    lb = ones((len(x), 1)) * -t

    for i in range(len(x)):
        zi2 = ri**2 - (x[i] - xc)**2 - (y[i] - yc)**2
        ze2 = re**2 - (x[i] - xc)**2 - (y[i] - yc)**2
        ub[i] = math.sqrt(ze2)
        if zi2 > 0.0:
            lb[i] = math.sqrt(zi2)

    return ub, lb


def dome_dub_dlb(x, y, thk, t, center=[5.0, 5.0], radius=5.0):
    """Update sensitivites of upper and lower bounds of the dome based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the dome
    t : float
        Parameter for lower bound in nodes in the boundary
    center : [float, float], optional
        x, y coordinates of the center of the dome, by default [5.0, 5.0]
    radius : float, optional
        The radius of the dome, by default 5.0

    Returns
    -------
    dub : array
        Values of the sensitivites of upper bound in the points
    dlb : array
        Values of the sensitivites of lower bound in the points
    """

    xc = center[0]
    yc = center[1]
    ri = radius - thk/2
    re = radius + thk/2
    dub = zeros((len(x), 1))
    dlb = zeros((len(x), 1))
    dubdx = zeros((len(x), len(x)))
    dubdy = zeros((len(x), len(x)))
    dlbdx = zeros((len(x), len(x)))
    dlbdy = zeros((len(x), len(x)))

    for i in range(len(x)):
        zi2 = ri**2 - (x[i] - xc)**2 - (y[i] - yc)**2
        ze2 = re**2 - (x[i] - xc)**2 - (y[i] - yc)**2
        ze = math.sqrt(ze2)
        dub[i] = 1/2 * re/ze
        dubdx[i, i] = 1/2/ze * -2 * (x[i] - xc)
        dubdy[i, i] = 1/2/ze * -2 * (y[i] - yc)
        if zi2 > 0.0:
            zi = math.sqrt(zi2)
            dlb[i] = -1/2 * ri/zi
            dlbdx[i, i] = 1/2/zi * -2 * (x[i] - xc)
            dlbdy[i, i] = 1/2/zi * -2 * (y[i] - yc)

    return dub, dlb, dubdx, dubdy, dlbdx, dlbdy


def dome_b_update(x, y, thk, fixed, center=[5.0, 5.0], radius=5.0):
    """Updates the ``b`` parameter of a dome for a given thickness

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the dome
    fixed : list
        List with indexes of the fixed vertices
    center : [float, float], optional
        x, y coordinates of the center of the dome, by default [5.0, 5.0]
    radius : float, optional
        The radius of the dome, by default 5.0

    Returns
    -------
    b : array
        The ``b`` parameter
    """

    [xc, yc] = center[:2]
    b = zeros((len(fixed), 2))

    for i in range(len(fixed)):
        i_ = fixed[i]
        theta = math.atan2((y[i_] - yc), (x[i_] - xc))
        x_ = abs(thk/2 * math.cos(theta))
        y_ = abs(thk/2 * math.sin(theta))
        b[i, :] = [x_, y_]

    return b


def dome_db(x, y, thk, fixed, center=[5.0, 5.0], radius=5.0):
    """Updates the ``db`` parameter of a dome for a given thickness

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the dome
    fixed : list
        List with indexes of the fixed vertices
    center : [float, float], optional
        x, y coordinates of the center of the dome, by default [5.0, 5.0]
    radius : float, optional
        The radius of the dome, by default 5.0

    Returns
    -------
    b : array
        The sensitivity of the ``b`` parameter
    """

    [xc, yc] = center[:2]
    db = zeros((len(fixed), 2))

    for i in range(len(fixed)):
        i_ = fixed[i]
        theta = math.atan2((y[i_] - yc), (x[i_] - xc))
        x_ = abs(1/2 * math.cos(theta))
        y_ = abs(1/2 * math.sin(theta))
        db[i, :] = [x_, y_]

    return db


def dome_b_update_with_n(x, y, n, fixed, b, center=[5.0, 5.0]):
    """Updates the ``b`` parameter of a dome for a given thickness considering as variable the offset from the surface

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    n : float
        Offset distance
    fixed : list
        List with indexes of the fixed vertices
    b : list
        Current ``b`` list
    center : [float, float], optional
        x, y coordinates of the center of the dome, by default [5.0, 5.0]

    Returns
    -------
    new_b : array
        The new ``b`` parameter
    """

    [xc, yc] = center[:2]
    new_b = zeros((len(fixed), 2))

    for i in range(len(fixed)):
        i_ = fixed[i]
        theta = math.atan2((y[i_] - yc), (x[i_] - xc))
        new_b[i, 0] = b[i, 0] - n * abs(math.cos(theta))
        new_b[i, 1] = b[i, 1] - n * abs(math.sin(theta))

    return new_b


def dome_db_with_n(x, y, fixed, center=[5.0, 5.0]):
    """Updates the sensitivites of the ``b`` parameter of a dome for a given thickness considering as variable the offset from the surface

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    n : float
        Offset distance
    fixed : list
        List with indexes of the fixed vertices
    b : list
        Current ``b`` list
    center : [float, float], optional
        x, y coordinates of the center of the dome, by default [5.0, 5.0]

    Returns
    -------
    db : array
        The sensitivity of the ``b`` parameter
    """

    [xc, yc] = center[:2]
    db = zeros((len(fixed), 2))

    for i in range(len(fixed)):
        i_ = fixed[i]
        theta = math.atan2((y[i_] - yc), (x[i_] - xc))
        db[i, 0] = - abs(math.cos(theta))
        db[i, 1] = - abs(math.sin(theta))

    return db



"""
********************************************************************************
compas_tno.shapes
********************************************************************************

.. currentmodule:: compas_tno.shapes

Shape
=====

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Shape

MeshDos
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    MeshDos

Shape topologies
================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    rectangular_topology

Circular Arch
=============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    arch_shape
    arch_shape_polar
    arch_ub_lb_update
    arch_dub_dlb
    arch_b_update
    arch_db

Pointed Arch
=============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    pointed_arch_shape
    pointed_arch_ub_lb_update
    pointed_arch_dub_dlb

Cross Vault
===========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    cross_vault_highfields_proxy
    cross_vault_highfields
    crossvault_ub_lb_update
    crossvault_middle_update
    crossvault_dub_dlb

Pointed Cross Vault
===================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    pointed_vault_heightfields_proxy
    pointed_vault_heightfields
    pointed_vault_middle_update
    pointed_vault_ub_lb_update
    pointed_vault_dub_dlb

Dome
====

.. autosummary::
    :toctree: generated/
    :nosignatures:

    dome_heightfields_proxy
    set_dome_heighfield
    set_dome_with_spr
    set_dome_polar_coord
    geom_dome
    dome_zt_update
    dome_ub_lb_update
    dome_dub_dlb
    dome_b_update
    dome_db
    dome_b_update_with_n
    dome_db_with_n

Pavillion Vault
===============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    pavillion_vault_highfields_proxy
    pavillion_vault_highfields
    pavillionvault_ub_lb_update
    pavillionvault_dub_dlb
    pavillionvault_b_update
    pavillionvault_db

Shells
================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    domical_vault
    parabolic_shell_highfields
    parabolic_shell_middle_update
    parabolic_shell_ub_lb_update

General Meshes
================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    general_ub_lb_update_with_t_middle_constant
    general_db_with_t_middle_constant
    general_ub_lb_update_with_t_middle_variable
    general_db_with_t_middle_variable
    general_ub_lb_update_with_t_intrados
    general_db_with_t_intrados
    general_ub_lb_update_with_s
    general_dub_dlb_with_s
    general_ub_lb_update_with_n
    general_dub_dlb_with_n

"""
from __future__ import absolute_import

from .meshdos import MeshDos
from .shape import Shape
from .rectangular_topology import rectangular_topology

import compas

if not compas.IPY:
    from .circular_arch import (
        arch_shape,
        arch_shape_polar,
        arch_ub_lb_update,
        arch_dub_dlb,
        arch_b_update,
        arch_db
    )
    from .crossvault import (
        cross_vault_highfields_proxy,
        cross_vault_highfields,
        crossvault_ub_lb_update,
        crossvault_middle_update,
        crossvault_dub_dlb,
    )
    from .dome import (
        dome_heightfields_proxy,
        set_dome_heighfield,
        set_dome_with_spr,
        set_dome_polar_coord,
        geom_dome,
        dome_zt_update,
        dome_ub_lb_update,
        dome_dub_dlb,
        dome_b_update,
        dome_db,
        dome_b_update_with_n,
        dome_db_with_n
    )
    from .general import (
        general_ub_lb_update_with_t_middle_constant,
        general_db_with_t_middle_constant,
        general_ub_lb_update_with_t_middle_variable,
        general_db_with_t_middle_variable,
        general_ub_lb_update_with_t_intrados,
        general_db_with_t_intrados,
        general_ub_lb_update_with_s,
        general_dub_dlb_with_s,
        general_ub_lb_update_with_n,
        general_dub_dlb_with_n
    )
    from .pavillionvault import (
        pavillion_vault_highfields_proxy,
        pavillion_vault_highfields,
        pavillionvault_ub_lb_update,
        pavillionvault_dub_dlb,
        pavillionvault_b_update,
        pavillionvault_db
    )
    from .pointed_arch import (
        pointed_arch_shape,
        pointed_arch_ub_lb_update,
        pointed_arch_dub_dlb
    )
    from .pointed_crossvault import (
        pointed_vault_heightfields_proxy,
        pointed_vault_heightfields,
        pointed_vault_middle_update,
        pointed_vault_ub_lb_update,
        pointed_vault_dub_dlb
    )
    from .shells import (
        domical_vault,
        parabolic_shell_highfields,
        parabolic_shell_middle_update,
        parabolic_shell_ub_lb_update
    )


__all__ = [
    'MeshDos',
    'Shape',

    'rectangular_topology',

    'arch_shape',
    'arch_shape_polar',
    'arch_ub_lb_update',
    'arch_dub_dlb',
    'arch_b_update',
    'arch_db',

    'cross_vault_highfields_proxy',
    'cross_vault_highfields',
    'crossvault_ub_lb_update',
    'crossvault_middle_update',
    'crossvault_dub_dlb',

    'dome_heightfields_proxy',
    'set_dome_heighfield',
    'set_dome_with_spr',
    'set_dome_polar_coord',
    'geom_dome',
    'dome_zt_update',
    'dome_ub_lb_update',
    'dome_dub_dlb',
    'dome_b_update',
    'dome_db',
    'dome_b_update_with_n',
    'dome_db_with_n',

    'general_ub_lb_update_with_t_middle_constant',
    'general_db_with_t_middle_constant',
    'general_ub_lb_update_with_t_middle_variable',
    'general_db_with_t_middle_variable',
    'general_ub_lb_update_with_t_intrados',
    'general_db_with_t_intrados',
    'general_ub_lb_update_with_s',
    'general_dub_dlb_with_s',
    'general_ub_lb_update_with_n',
    'general_dub_dlb_with_n',

    'pavillion_vault_highfields_proxy',
    'pavillion_vault_highfields',
    'pavillionvault_ub_lb_update',
    'pavillionvault_dub_dlb',
    'pavillionvault_b_update',
    'pavillionvault_db',

    'pointed_arch_shape',
    'pointed_arch_ub_lb_update',
    'pointed_arch_dub_dlb',

    'pointed_vault_heightfields_proxy',
    'pointed_vault_heightfields',
    'pointed_vault_middle_update',
    'pointed_vault_ub_lb_update',
    'pointed_vault_dub_dlb',

    'domical_vault',
    'parabolic_shell_highfields',
    'parabolic_shell_middle_update',
    'parabolic_shell_ub_lb_update'
]



"""
********************************************************************************
compas_tno.viewers
********************************************************************************

.. currentmodule:: compas_tno.viewers

Viewer
========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Viewer


Annimation
==========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    animation_from_optimisation
    animation_from_section

"""

from __future__ import absolute_import

from compas_view2.objects import Object
from compas_view2.objects import MeshObject
from compas_tno.shapes.meshdos import MeshDos
from compas_tno.diagrams.form import FormDiagram
from compas_tno.diagrams.force import ForceDiagram

from .animation import (
    animation_from_optimisation,
    animation_from_section
)
from .viewer import Viewer

Object.register(MeshDos, MeshObject)
Object.register(FormDiagram, MeshObject)
Object.register(ForceDiagram, MeshObject)


__all__ = [
    'Viewer',
    'animation_from_optimisation',
    'animation_from_section'
]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import Rhino
from Rhino.Geometry import Point3d

import compas_rhino

from compas.geometry import scale_vector
from compas.geometry import add_vectors
from compas.geometry import subtract_vectors

from compas_ui.objects import MeshObject


__all__ = ['DiagramObject']


class DiagramObject(MeshObject):
    """A diagram object represents a form or force diagram in the Rhino view.
    """

    @property
    def diagram(self):
        """The diagram associated with the object."""
        return self.mesh

    @diagram.setter
    def diagram(self, diagram):
        self.mesh = diagram

    def unselect(self):
        """Unselect all Rhino objects associated with this diagram object."""
        compas_rhino.rs.UnselectObjects(self.artist.guids)

    def select_vertex(self, message="Select Vertex."):
        """Manually select one vertex in the Rhino view.

        Returns
        -------
        int
            The identifier of the selected vertex.
        """
        pointfilter = compas_rhino.rs.filter.point
        guid = compas_rhino.rs.GetObject(message=message, preselect=True, select=True, filter=pointfilter)
        if guid and guid in self.guid_vertex:
            return self.guid_vertex[guid]

    def select_edge(self, message="Select Edge."):
        """Manually select one edge in the Rhino view.

        Returns
        -------
        tuple of int
            The identifier of the selected edge.
        """
        curvefilter = compas_rhino.rs.filter.curve
        guid = compas_rhino.rs.GetObject(message=message, preselect=True, select=True, filter=curvefilter)
        if guid and guid in self.guid_edge:
            return self.guid_edge[guid]

    def move(self):
        """Move the entire diagram.

        Returns
        -------
        bool
            True if the operation is successful.
            False otherwise.
        """
        def OnDynamicDraw(sender, e):
            translation = subtract_vectors(list(e.CurrentPoint), start)
            for vertex in vertex_xyz:
                vertex_xyz[vertex] = add_vectors(vertex_xyz0[vertex], translation)
            for u, v in iter(edges):
                e.Display.DrawDottedLine(Point3d(* vertex_xyz[u]), Point3d(* vertex_xyz[v]), color)

        color = Rhino.ApplicationSettings.AppearanceSettings.FeedbackColor
        vertex_xyz = self.artist.vertex_xyz
        vertex_xyz0 = {vertex: xyz[:] for vertex, xyz in vertex_xyz.items()}
        edges = list(self.diagram.edges())
        start = compas_rhino.pick_point('Point to move from?')
        if not start:
            return False
        start = list(start)

        gp = Rhino.Input.Custom.GetPoint()
        gp.SetCommandPrompt('Point to move to?')
        gp.DynamicDraw += OnDynamicDraw
        gp.Get()
        if gp.CommandResult() != Rhino.Commands.Result.Success:
            return False

        end = list(gp.Point())
        translation = subtract_vectors(end, start)
        self.location = add_vectors(self.location, translation)
        return True

    def move_vertex(self, vertex, constraint=None, allow_off=None):
        """Move one vertex of the diagram and update the data structure to the new geometry.

        Parameters
        ----------
        vertex : int
            The identifier of the vertex.

        Other Parameters
        ----------------
        constraint : :class:`Rhino.Geometry.GeometryBase`, optional
        allow_off : bool, optional

        Returns
        -------
        bool
            True if the operation was successful.
            False otherwise.
        """
        def OnDynamicDraw(sender, e):
            sp = e.CurrentPoint
            for ep in nbrs:
                e.Display.DrawDottedLine(sp, ep, color)

        color = Rhino.ApplicationSettings.AppearanceSettings.FeedbackColor
        diagram = self.diagram
        vertex_xyz = self.artist.vertex_xyz

        if '_is_edge' in diagram.default_edge_attributes:
            nbrs = [Point3d(* vertex_xyz[nbr]) for nbr in diagram.vertex_neighbors(vertex) if diagram.edge_attribute((vertex, nbr), '_is_edge')]
        else:
            nbrs = [Point3d(* vertex_xyz[nbr]) for nbr in diagram.vertex_neighbors(vertex)]

        gp = Rhino.Input.Custom.GetPoint()
        gp.SetCommandPrompt('Point to move to?')
        if constraint:
            gp.Constrain(constraint, allow_off)
        gp.DynamicDraw += OnDynamicDraw
        gp.Get()
        if gp.CommandResult() != Rhino.Commands.Result.Success:
            return False

        point = list(gp.Point())
        xyz0 = vertex_xyz[vertex]
        dxyz0 = subtract_vectors(point, xyz0)
        dxyz = scale_vector(dxyz0, 1 / self.scale)
        xyz = diagram.vertex_attributes(vertex, 'xyz')
        xyz[:] = add_vectors(xyz, dxyz)
        diagram.vertex_attributes(vertex, 'xyz', xyz)
        return True

    def move_vertices(self, vertices):
        """Move a selection of vertices of the diagram and update the data structure to the new geometry.

        Parameters
        ----------
        vertices : list
            The identifiers of the vertices.

        Returns
        -------
        bool
            True if the operation was successful.
            False otherwise.
        """
        def OnDynamicDraw(sender, e):
            end = e.CurrentPoint
            vector = end - start
            for a, b in lines:
                e.Display.DrawDottedLine(a + vector, b + vector, color)
            for a, b in connectors:
                e.Display.DrawDottedLine(a + vector, b, color)

        diagram = self.diagram
        vertex_xyz = self.artist.vertex_xyz

        color = Rhino.ApplicationSettings.AppearanceSettings.FeedbackColor
        lines = []
        connectors = []
        for vertex in vertices:
            a = vertex_xyz[vertex]
            nbrs = diagram.vertex_neighbors(vertex)
            for nbr in nbrs:
                if '_is_edge' in diagram.default_edge_attributes and not diagram.edge_attribute((vertex, nbr), '_is_edge'):
                    continue
                b = vertex_xyz[nbr]
                line = [Point3d(* a), Point3d(* b)]
                if nbr in vertices:
                    lines.append(line)
                else:
                    connectors.append(line)

        gp = Rhino.Input.Custom.GetPoint()

        gp.SetCommandPrompt('Point to move from?')
        gp.Get()
        if gp.CommandResult() != Rhino.Commands.Result.Success:
            return False
        start = gp.Point()

        gp.SetCommandPrompt('Point to move to?')
        gp.SetBasePoint(start, False)
        gp.DrawLineFromPoint(start, True)
        gp.DynamicDraw += OnDynamicDraw
        gp.Get()
        if gp.CommandResult() != Rhino.Commands.Result.Success:
            return False
        end = gp.Point()

        dxyz0 = list(end - start)
        dxyz = scale_vector(dxyz0, 1 / self.scale)
        for vertex in vertices:
            xyz = diagram.vertex_attributes(vertex, 'xyz')
            xyz[:] = add_vectors(xyz, dxyz)
            diagram.vertex_attributes(vertex, 'xyz', xyz)
        return True


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino
from compas_tno.optimisers import Optimiser


__commandname__ = "TNO_optimization_form"


def RunCommand(is_interactive):

    if 'TNO' not in sc.sticky:
        compas_rhino.display_message('TNO has not been initialised yet.')
        return

    scene = sc.sticky['TNO']['scene']

    data = {}

    data['printout'] = True

    # objective
    obj = compas_rhino.rs.GetString("Optimisation Objective", "min", ["min", "max", "t", "Cancel"])
    if not obj or obj == "Cancel":
        return
    data['objective'] = obj

    # variables
    options = ['q', 'zb', 'xyb', 't']
    items = ("ForceDensitiesQ", "False", "True"), ("HeightSupportsZb", "False", "True"), ("PlanarSupportsXyb", "False", "True"), ("Thickness", "False", "True")
    results = compas_rhino.rs.GetBoolean("Select Variables", items, (True, True, False, False))
    if not results:
        return
    data['variables'] = [i for indx, i in enumerate(options) if results[indx]]

    # constraints
    options = ['funicular', 'envelope', 'reac_bounds', 'envelopexy']
    items = ("Funicular", "False", "True"), ("Envelope", "False", "True"), ("ReactionBounds", "False", "True"), ("EnvelopeXy", "False", "True")
    results = compas_rhino.rs.GetBoolean("Select Constraints", items, (True, True, False, False))
    if not results:
        return
    data['constraints'] = [i for indx, i in enumerate(options) if results[indx]]

    # features
    options = ['fixed', 'sym']
    items = ("FixedProjection", "False", "True"), ("Symmetry", "False", "True")
    results = compas_rhino.rs.GetBoolean("Select Features", items, (True, False))
    if not results:
        return
    data['features'] = [i for indx, i in enumerate(options) if results[indx]]

    # solver
    lib = compas_rhino.rs.GetString("Select Library", "SLSQP", ["SLSQP", "IPOPT", "Cancel"])
    if not lib or lib == "Cancel":
        return
    data['library'] = lib
    data['solver'] = lib

    # derivatives
    deriv = compas_rhino.rs.GetString("Use analytical derivatives", "True", ["True", "False", "Cancel"])
    if deriv == "True":
        data['gradient'] = True
        data['jacobian'] = True

    # max iter
    max_iter = compas_rhino.rs.GetInteger("Maximum of Iterations", 500)
    if not max_iter:
        return
    data['max_iter'] = max_iter

    # starting point
    sp = compas_rhino.rs.GetString("Select Starting point", "current", ["current", "loadpath", "ParalelliseTna", "Cancel"])
    if not sp or sp == "Cancel":
        return
    data['starting_point'] = sp

    # qmax qmin
    qmin = compas_rhino.rs.GetReal("Assign qmin", -1e+4)
    if not qmin:
        pass
    else:
        data['qmin'] = float(qmin)
    qmax = compas_rhino.rs.GetReal("Assign qmax", +1e-8)
    if not qmax:
        pass
    else:
        data['qmax'] = float(qmax)

    objects = scene.find_by_name('Optimiser')
    if not objects:
        optimiser = Optimiser()
        optimiser.settings = data
        scene.add(optimiser, name='Optimiser', layer=None)
        objects = scene.find_by_name('Optimiser')
        optimiserobject = objects[0]
    else:
        optimiserobject = objects[0]
        optimiser = optimiserobject.optimiser
        optimiser.settings = data

    optimiserobject.update_object_from_optimiser()

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================
if __name__ == '__main__':

    RunCommand(True)



from compas_tno.shapes import Shape
from compas_tno.diagrams import FormDiagram
from compas_tno.viewers import Viewer
from compas_tno.analysis import Analysis
from compas_tno.plotters import TNOPlotter
from compas.geometry import distance_point_point_xy
from compas_tno.utilities import form_add_lines_support
from numpy import zeros
from compas_view2.objects import Arrow
from compas.colors import Color


# ----------------------------------------
# 1. Shape geometric definition
# ----------------------------------------
spr_angle = 30.0
L = 10.0
thk = 0.50
xy_span = [[0, L], [0, L]]
vault = Shape.create_crossvault(xy_span=xy_span, thk=thk, spr_angle=30)

# ------------------------------------------------------------------------
# 2. Form diagram geometric definition with additional line to supports
# ------------------------------------------------------------------------
discretisation = 16
form = FormDiagram.create_cross_form(xy_span=xy_span, discretisation=discretisation)

load_pos = 3
xc = yc = L/2
yp = 2.5
yp = yc - load_pos/(discretisation/2) * yc
for key in form.vertices():
    pt = form.vertex_coordinates(key)
    if distance_point_point_xy(pt, [xc, yp, 0.0]) < 1e-3:
        loaded_node = key
        break

supports = []
for key in form.vertices_where({'is_fixed': True}):
    x, y, z = form.vertex_coordinates(key)
    if y < yc:
        supports.append(key)

print(loaded_node, supports)

form, loaded_node = form_add_lines_support(form, loaded_node=loaded_node, supports=supports)

plotter = TNOPlotter(form)
plotter.draw_form(scale_width=False, color=Color.black())
plotter.draw_supports(color=Color.red())
plotter.show()

# ------------------------------------------------------------------------
# 4. Define applied load case
# ------------------------------------------------------------------------

n = form.number_of_vertices()
load_direction = zeros((n, 1))
load_direction[loaded_node] = -1.0
print('New Loaded Node:', loaded_node)

# --------------------------------------------
# 5. Maximum load problem and visualisation
# --------------------------------------------
analysis = Analysis.create_max_load_analysis(form, vault, load_direction=load_direction, max_lambd=300, printout=True)
analysis.apply_selfweight()
analysis.apply_envelope()
analysis.set_up_optimiser()
analysis.run()

viewer = Viewer(form)
viewer.settings['scale.reactions'] = 0.004
viewer.draw_thrust()
viewer.draw_cracks()
viewer.draw_shape()
viewer.draw_reactions()

length = 2.0
x, y, z = form.vertex_coordinates(loaded_node)
z += length + 0.1
arrow = Arrow([x, y, z], [0, 0, -length])
viewer.app.add(arrow, linecolor=(0, 0, 0), facecolor=(0, 0, 0))

viewer.show()
